// $imfile.cpp 1.5 milbo$ routines for image files

#include "all.hpp"

static char sgErr[1000];    // place to store error messages

//-----------------------------------------------------------------------------
// Load a .bmp file.  See the header of sLoadImage below for some details.

static char *
sLoadBmp (Image *pImg, RgbImage *pRgbImg,                           // out
        const char *sFile, bool fExitOnErr, bool fRemovePad=true,   // in
        bool fDeleteOnErr=false)                                    // in
{
BITMAPFILEHEADER    bmfHeader;
BITMAPINFOHEADER    bmiHeader;
FILE                *pBmpFile;
int width = 0;
int height = 0;
unsigned Len = 0;
byte *pRgbTemp = NULL;
int nPad = 0;
sgErr[0] = 0;
pBmpFile = fopen(sFile, "rb");
if (!pBmpFile)
    sprintf(sgErr, "Can't open %s", sFile);

if (!sgErr[0] &&
        ((Len = fread(&bmfHeader, 1, sizeof(bmfHeader), pBmpFile)) != sizeof(bmfHeader) ||
            fread(&bmiHeader, 1, sizeof(bmiHeader), pBmpFile) != sizeof(bmiHeader)))
    sprintf(sgErr, "Can't read %s%s", sFile, (Len? " (too short to be a BMP file)": ""));

if (!sgErr[0] && (bmfHeader.bfType != ((WORD) ('M' << 8) | 'B')))
    sprintf(sgErr, "%s is not a BMP file (first two bytes are not \"BM\")", sFile);

if (!sgErr[0] &&
        (bmiHeader.biBitCount!=24 || bmiHeader.biClrUsed || bmiHeader.biCompression))
    {
    sprintf(sgErr, "%s is not a windows 24 bit uncompressed RGB file ", sFile);
    if (bmiHeader.biBitCount!=24)
        strcat(sgErr, "(bit count is not 24)");
    else if (bmiHeader.biClrUsed)
        strcat(sgErr, "(ClrUsed is set)");
    else if (bmiHeader.biCompression)
        strcat(sgErr, "(file is compressed)");
    }
if (!sgErr[0])
    {
    width = bmiHeader.biWidth;
    height = bmiHeader.biHeight;
    if (width < 1 || width > MAX_IMAGE_DIM || height > MAX_IMAGE_DIM || height < 1)
        sprintf(sgErr, "%s has a strange width %d or height %d", sFile, width, height);
    }
if (!sgErr[0])
    {
    nPad = width % 4;
    Len = (((3 * width) + nPad) * height);
    pRgbTemp = (byte *)malloc(Len); // have to use a temporary buffer because of pad
    if (fread(pRgbTemp, 1, Len, pBmpFile) != Len)
        {
        free(pRgbTemp);
        sprintf(sgErr, "Can't read %s (it's shorter than its header says it should be)",
            sFile);
        }
    }
if (pBmpFile)
    fclose(pBmpFile);
if (sgErr[0])
    {
    if (fDeleteOnErr)
        unlink(sFile);
    if (fExitOnErr)
        Err(sgErr);
    else
        return sgErr;
    }
if (pImg)
    {
    // copy pRgbTemp into pImg, converting to gray and discarding the pad

    pImg->dim(width, height);
    byte *pTo = pImg->buf;
    for (int iy = height - 1; iy >= 0; iy--)
        {
        byte *pFrom = pRgbTemp + (3 * iy * width) + (nPad * iy);
        for (int ix = 0; ix < width; ix++)
            {
            tRGB Rgb = *(tRGB *)pFrom;
            *pTo++ = (byte)RgbToGray(Rgb);  // CIE conversion to gray
            pFrom += 3;
            }
        }
    }
if (pRgbImg)
    {
    // copy pRgbTemp into pRgbImg, discarding the nasty little pad

    pRgbImg->dim(width + (fRemovePad? 0: nPad), height);
    tRGB *pTo = pRgbImg->buf;
    tRGB *pFrom = (tRGB *)pRgbTemp;
    for (int iy = 0; iy < height; iy++)
        {
        for (int ix = 0; ix < width; ix++)
            *pTo++ = *pFrom++;
        byte *r = (byte *)pFrom;
        r += nPad;
        pFrom = (tRGB *)r;
        }
    }
free(pRgbTemp);

return NULL;    // success
}

//-----------------------------------------------------------------------------
// Load a .pgm file.  See the header of sLoadImage for some details.

static const char *
sLoadPgm (Image *pImg, RgbImage *pRgbImg,           // out
           const char *sFile, bool fExitOnErr)      // in
{
FILE *pFile = fopen(sFile, "rb");
if (!pFile)
    if (fExitOnErr)
        Err("Can't open %s", sFile);
    else
        {
        sprintf(sgErr, "Can't open %s", sFile);
        return sgErr;
        }
int width = -1, height = -1, DataSize, nTokens = 0;
do
    {
    char sLine[SLEN];
    fgets(sLine, SLEN, pFile);
    char *comment = strchr(sLine,'#');
    if (comment != NULL)
        *comment = 0;
    char *token = NULL, *input = &sLine[0];
    while (nTokens < 4 && (token = strtok(input," \t\n\r")) != NULL)
        {
        switch (nTokens)
            {
            case 0:
                if (strcmp(token,"P5") != 0)
                    if (fExitOnErr)
                        Err("%s has a bad header (first two bytes are not \"P5\")", sFile);
                    else
                        return "bad header (first two bytes are not \"P5\")";
                break;
            case 1:
                width = atoi(token);
                break;
            case 2:
                height = atoi(token);
                break;
            case 3:
                DataSize = atoi(token);
                if (DataSize != 255)
                    if (fExitOnErr)
                        Err("This program can only deal with data sizes of 255 and "
                            "%s has a data size of %d", sFile, DataSize);
                    else
                        return "data size not 255";
                break;
            default:
                break;
            }
        nTokens++;
        input = NULL;
        }
    }
while (nTokens < 4);

if (width < 1 || width > MAX_IMAGE_DIM || height > MAX_IMAGE_DIM || height < 1)
    if (fExitOnErr)
        Err("%s has a strange width %d or height %d", sFile, width, height);
    else
        return "strange width or height";

Image ImgTemp(width, height);
Fread((char *)ImgTemp.buf , 1, width * height, pFile, sFile);
fclose(pFile);

if (pRgbImg)
    {
    pRgbImg->dim(width, height);
    byte *pTo = (byte *)(pRgbImg->buf);
    for (int iy = height-1; iy >= 0; iy--)
        for (int ix = 0; ix < width; ix++)
            {
            byte b = ImgTemp(ix, iy);
            *pTo++ = b;
            *pTo++ = b;
            *pTo++ = b;
            }
    }
if (pImg)
    *pImg = ImgTemp;

return NULL;    // success
}

//-----------------------------------------------------------------------------
const char *
sLoadPgm(Image &Img,                            // out
           const char *sFile, bool fExitOnErr)  // in
{
return sLoadPgm(&Img, NULL, sFile, fExitOnErr);
}

//-----------------------------------------------------------------------------
// Load a .ppm file.  See the header of sLoadImage for some details.

static const char *
sLoadPpm (Image *pImg, RgbImage *pRgbImg, const char *sFile, bool fExitOnErr)
{
FILE *pFile = fopen(sFile, "rb");
if (!pFile)
    if (fExitOnErr)
        Err("Can't open %s", sFile);
    else
        {
        sprintf(sgErr, "Can't open %s", sFile);
        return sgErr;
        }
int width = -1, height = -1, DataSize, nTokens = 0;
do
    {
    char sLine[SLEN];
    fgets(sLine, SLEN, pFile);
    char *comment = strchr(sLine,'#');
    if (comment != NULL)
        *comment = 0;
    char *token = NULL, *input = &sLine[0];
    while (nTokens < 4 && (token = strtok(input," \t\n\r")) != NULL)
        {
        switch (nTokens)
            {
            case 0:
                if (strcmp(token,"P6") != 0)
                    if (fExitOnErr)
                        Err("%s has a bad header (first two bytes are not \"P6\")", sFile);
                    else
                        return "bad header (first two bytes are not \"P6\")";
                break;
            case 1:
                width = atoi(token);
                break;
            case 2:
                height = atoi(token);
                break;
            case 3:
                DataSize = atoi(token);
                if (DataSize != 255)
                    if (fExitOnErr)
                        Err("This program can only deal with data sizes of 255\n"
                            "and %s has a data size of %d", sFile, DataSize);
                    else
                        return "data size not 255";
                break;
            default:
                break;
            }
        nTokens++;
        input = NULL;
        }
    }
while (nTokens < 4);
if (width < 1 || width > MAX_IMAGE_DIM || height > MAX_IMAGE_DIM || height < 1)
    if (fExitOnErr)
        Err("%s has a strange width %d or height", sFile, width, height);
    else
        return "strange width or height";
RgbImage RgbTemp(width, height);
Fread(RgbTemp.buf, 1, 3 * width * height, pFile, sFile);
fclose(pFile);

if (pImg)
    {
    // copy pRgbImg into buf, converting to gray

    pImg->dim(width, height);
    for (int iy = 0; iy < height; iy++)
        for (int ix = 0; ix < width; ix++)
            {
            // CIE conversion to gray, add 500 to take care of rounding.
            // Colors are reversed so have to take care of that too.
            tRGB c = RgbTemp(ix, iy);
            (*pImg)(ix, iy) = (byte)
                ((299 * c.Blue + 587 * c.Green + 114 * c.Red + 500)/1000);
            }
    }
if (pRgbImg)
    {
    // Image comes in upside down with colors reversed.  Fix that.

    pRgbImg->dim(width, height);
    for (int iy = 0, y1 = height-1; iy < height; iy++, y1--)
        for (int ix = 0; ix < width; ix++)
            {
            (*pRgbImg)(ix, iy).Red = RgbTemp(ix, y1).Blue;
            (*pRgbImg)(ix, iy).Green = RgbTemp(ix, y1).Green;
            (*pRgbImg)(ix, iy).Blue = RgbTemp(ix, y1).Red;
            }
    }
return NULL;    // success
}

//-----------------------------------------------------------------------------
static FILE *
pOpenTempFile (char sFile[],                                    // out
                const char sTemplate[], const char sMode[])     // in
{
char sTemplate1[SLEN];
sprintf(sTemplate1, "%s_XXXXXX", sTemplate);

char *sBase = _mktemp(sTemplate1);
if (!sBase)
    Err("Can't generate temporary filename");

char *sTmpDir = getenv("TMPDIR");
if (!sTmpDir)
    sTmpDir = getenv("TEMP");
if (!sTmpDir)
    sTmpDir = getenv("TMP");
if (!sTmpDir)
    Err("Can't find temporary directory (no environment variables TMPDIR TEMP TMP)");

_makepath(sFile, "", sTmpDir, sBase, "bmp");

return Fopen(sFile, sMode);     // will give err msg and exit if can't open the file
}

//-----------------------------------------------------------------------------
// Load an image.  Knows about different types of files like BMPs and JPEGs
//
// Returns an error message if necessary, else NULL.
//
// This uses the file extension to determine file type.
//
// pImg is gray version of the image.      Caller must free.  Set to NULL if don't need.
// pRgbImg is color version of the image.  Caller must free.  Set to NULL if don't need.
//
// sFile is the file name.
// Set fVerbose if you want to tell the user that you are reading the file.
//
// fRemovePad=false only applies to RGB bitmaps and is needed
// for StretchDIBits() under windows

const char *
sLoadImage (Image *pImg, RgbImage *pRgbImg,                                     // out
            const char *sFile, bool fVerbose, bool fExitOnErr, bool fRemovePad) // in
{
if (fVerbose)
    lprintf("Reading %s\n", sFile);
if (sFile[0] == 0)
    {
    sprintf(sgErr, "Null path name, can't load image");
    if (fExitOnErr)
        Err(sgErr);
    else
        return sgErr;
    }
char sBaseName[_MAX_FNAME], sExt[_MAX_EXT];
_splitpath(sFile, NULL, NULL, sBaseName, sExt);

if (sExt[0] == 0)
    {
    sprintf(sgErr,
        "Image file %s has no extension: can't determine the image type to read from disk",
        sFile);
    if (fExitOnErr)
        Err(sgErr);
    else
        return sgErr;
    }
switch (toupper(sExt[1]))
    {
    case 'B':
        return sLoadBmp(pImg, pRgbImg, sFile, fExitOnErr, fRemovePad);
        break;
    case 'J':
        {
        // File is a JPEG file: create a temporary BMP file and use that instead.
        // TODO this isn't an efficient way of doing it.

        char sTempPath[SLEN];
        FILE *pTempFile = pOpenTempFile(sTempPath, sBaseName, "w+b");
        ConvertJpgFileToBmpFile(sFile, pTempFile, 0);
        fclose(pTempFile);
        char *s = sLoadBmp(pImg, pRgbImg, sTempPath, fExitOnErr, true /* fDeleteOnErr */);
        unlink(sTempPath);
        return s;
        }
        break;
    case 'P':
        if (toupper(sExt[2]) == 'G')
            return sLoadPgm(pImg, pRgbImg, sFile, fExitOnErr);
        else
            return sLoadPpm(pImg, pRgbImg, sFile, fExitOnErr);
        break;
    default:
        {
        // open merely so we can give a decent error message
        FILE *pFile = fopen(sFile, "rb");
        if (!pFile)
            {
            sprintf(sgErr, "Can't open %s", sFile);
            if (fExitOnErr)
                Err(sgErr);
            else
                return sgErr;
            }
        fclose(pFile);
        sprintf(sgErr, "%s is an unrecognized image type", sFile);
        if (fExitOnErr)
            Err(sgErr);
        else
            return sgErr;
        break;
        }
    }
return NULL;    // success
}

//-----------------------------------------------------------------------------
// Load an image.  Knows about different types of files like BMPs and JPEGs.
// This uses the file extension to determine file type.
//
// Img is the gray image
// sFile is the file name
// Set fVerbose if you want to tell the user that you are reading the file

const char *
sLoadImage (Image &Img,                                             // out
              const char *sFile, bool fVerbose, bool fExitOnErr)    // in
{
return sLoadImage(&Img, NULL, sFile, fVerbose, fExitOnErr, true);
}

//-----------------------------------------------------------------------------
// See above header.  RgbImg is the color image.
// fRemovePad=false is needed to for StretchDIBits() under windows

const char *
sLoadImage (RgbImage &RgbImg,                                                   // out
            const char *sFile, bool fVerbose, bool fExitOnErr, bool fRemovePad) // in
{
return sLoadImage(NULL, &RgbImg, sFile, fVerbose, fExitOnErr, fRemovePad);
}

//-----------------------------------------------------------------------------
static void
WriteBmpHeader (FILE *pBmpFile, char sFile[], int width, int height)    // in: all
{
BITMAPFILEHEADER    bmfHeader;
BITMAPINFOHEADER    bmiHeader;

ASSERT(width > 0 && height > 0);
int nPad = width % 4;
int nNewWidth = width * 3 + nPad;

bmfHeader.bfType        = ((WORD) ('M' << 8) | 'B');
bmfHeader.bfSize        = (height * nNewWidth) + 56;
bmfHeader.bfReserved1   = 0;
bmfHeader.bfReserved2   = 0;
bmfHeader.bfOffBits     = 54;

Fwrite(&bmfHeader, 1, sizeof(bmfHeader), pBmpFile, sFile);

bmiHeader.biSize            = 40;       // sizeof(bmiHeader) always 40
bmiHeader.biWidth           = width;    // width
bmiHeader.biHeight          = height;   // height
bmiHeader.biPlanes          = 1;        // 1    These numbers are for RGB
bmiHeader.biBitCount        = 24;       // 24   24 bit uncompressed files
bmiHeader.biCompression     = 0;        // 0 no compression
bmiHeader.biSizeImage       = 0;        // 0
bmiHeader.biXPelsPerMeter   = 2834;     // 2834 magic number copied from existing bitmaps
bmiHeader.biYPelsPerMeter   = 2834;     // 2834 ditto
bmiHeader.biClrUsed         = 0;        // 0 no color map entries
bmiHeader.biClrImportant    = 0;        // 0

Fwrite(&bmiHeader, 1, sizeof(bmiHeader), pBmpFile, sFile);
}

//-----------------------------------------------------------------------------
static void
WriteBmpPixels (FILE *pBmpFile, char sFile[],           // in: all
                byte *buf, int width, int height)
{
int nPad = width % 4;
int Pixel;
for (int iy = height - 1; iy >= 0; iy--)
    {
    for (int ix = 0; ix < width; ix++)
        {
        Pixel = *(buf + ix + iy * width);
        Pixel = (Pixel | (Pixel << 8) | (Pixel << 16));     // gray to RGB
        Fwrite(&Pixel, 1, 3, pBmpFile, sFile);
        }
    if (nPad)
        {
        Pixel = 0;
        Fwrite(&Pixel, 1, nPad, pBmpFile, sFile);
        }
    }
Pixel = 0;
Fwrite(&Pixel, 1, 2, pBmpFile, sFile);  // photoshop puts 2 bytes on end, so we do too
}

//-----------------------------------------------------------------------------
static void
WriteBmpPixels (FILE *pBmpFile, char sFile[],           // in: all
                tRGB *buf, int width, int height)
{
int nPad = width % 4;
int PadPixel = 0;
for (int iy = 0; iy < height; iy++)
    {
    Fwrite(buf + iy * width, 1, 3 * width, pBmpFile, sFile);
    if (nPad)
        Fwrite(&PadPixel, 1, nPad, pBmpFile, sFile);
    }
Fwrite(&PadPixel, 1, 2, pBmpFile, sFile);
}

//-----------------------------------------------------------------------------
// write the given buffer as a 24 bit uncompressed windows BMP file

void
WriteBmp (const Image &Img, const char sFile[], bool fVerbose)  // in: all
{
FILE    *pBmpFile;
char    sDrive[_MAX_DRIVE], sDir[_MAX_DIR], sBaseName[_MAX_FNAME], sExt[_MAX_EXT];
char    sFile1[SLEN];
int width = Img.width, height = Img.height;

_splitpath(sFile, sDrive, sDir, sBaseName, sExt); // make sure file has a .bmp extension

// if user has strange extension don't silently overwrite it (4 is ".bmp")
ASSERT(strlen(sExt) <= 4);

_makepath(sFile1, sDrive, sDir, sBaseName, ".bmp");

if (fVerbose)
    lprintf("Writing %s\n", sFile1);

pBmpFile = Fopen(sFile1, "wb");     // will invoke Err if can't open
WriteBmpHeader(pBmpFile, sFile1, width, height);
WriteBmpPixels(pBmpFile, sFile1, Img.buf, width, height);
fclose(pBmpFile);
}

//-----------------------------------------------------------------------------
void
WriteBmp (const RgbImage &Img, const char sFile[], bool fVerbose)   // in: all
{
FILE    *pBmpFile;
char    sDrive[_MAX_DRIVE], sDir[_MAX_DIR], sBaseName[_MAX_FNAME], sExt[_MAX_EXT];
char    sFile1[SLEN];
int width = Img.width, height = Img.height;

_splitpath(sFile, sDrive, sDir, sBaseName, sExt);  // make sure file has a .bmp extension

// if user has strange extension don't silently overwrite it (4 is ".bmp")
ASSERT(strlen(sExt) <= 4);

_makepath(sFile1, sDrive, sDir, sBaseName, ".bmp");

if (fVerbose)
    lprintf("Writing %s %dx%d\n", sFile1, width, height);

pBmpFile = Fopen(sFile1, "wb");
WriteBmpHeader(pBmpFile, sFile1, width, height);
WriteBmpPixels(pBmpFile, sFile1, Img.buf, width, height);
fclose(pBmpFile);
}
