// $mat.cpp 1.5 milbo$ matrix classes, matrix elements are doubles
// Initially derived from gslwrap-0.2\matrix_double.cc
//
// Note: this has been stripped down for the "stasm" build.
//       Refer to the original mat.cpp if you want more
//       functionality (but more mess).

#include "all.hpp"

static const bool USE_EQ_DOUBLES = true;        // 1 to for match within tolerance
static const char *MAT_PRINT_FORMAT = "% 8.2f"; // default mat print format

namespace GslMat
{
//-----------------------------------------------------------------------------
Mat::Mat (const double Vals[], size_t nrows1, size_t ncols1)
{
m = gsl_matrix_alloc(nrows1, ncols1);
size_t iVal = 0;
for (size_t iRow = 0; iRow < nrows1; iRow++)
    for (size_t iCol = 0; iCol < ncols1; iCol++)
        gsl_matrix_set(m, iRow, iCol, Vals[iVal++]);
}

//-----------------------------------------------------------------------------
// declare matrix as vector and init
Mat::Mat (const double Vals[], size_t nelems, const char sIsRowVec[])
{
size_t iVal = 0;
if (sIsRowVec)                              // create row vector if ANY string
    {                                       // column matrix: one column, many rows
    m = gsl_matrix_alloc(1, nelems);
    for (size_t iCol = 0; iCol < nelems; iCol++)
        gsl_matrix_set(m, 0, iCol, Vals[iVal++]);
    }
else                                        // row matrix: one row, many columns
    {
    m = gsl_matrix_alloc(nelems, 1);
    for (size_t iRow = 0; iRow < nelems; iRow++)
        gsl_matrix_set(m, iRow, 0, Vals[iVal++]);
    }
}

//-----------------------------------------------------------------------------
void Mat::getDim (int *pnrows, int *pncols) const
{
if (m)
    {
    *pnrows = nrows();
    *pncols = ncols();
    }
else
    {
    *pnrows = 0;
    *pncols = 0;
    }
}

//-----------------------------------------------------------------------------
// If dimensions are same then matrix not touched.
// Else redimensions the matrix.  The new contents are unspecified.
// This function is quicker than dimKeep() or dimClear()

void Mat::dim (size_t nrows1, size_t ncols1)
{
if (nrows1 == 0 || ncols1 == 0) // zero size?
    {
    if (m)
        {
        ASSERT(m->owner);       // only the owner can change the size (a view can't)
        gsl_matrix_free(m);
        m = NULL;
        }
    }
else if (!m)
    m = gsl_matrix_alloc(nrows1, ncols1);
else if (nrows() != nrows1 || ncols() != ncols1)
    {
    ASSERT(m->owner);
    gsl_matrix_free(m);
    m = gsl_matrix_alloc(nrows1, ncols1);
    }
}

void Mat::dim (size_t nelems, const char *sIsRowVec) // redimension, new contents unspecified
{
if (sIsRowVec)                      // create row vector if ANY string
    dim(1, nelems);
else
    dim(nelems, 1);
}

void Mat::dim (const Mat &other)    // make this the same size as other
{
dim(other.nrows(), other.ncols());
}

//-----------------------------------------------------------------------------
// If dimensions are same then matrix not touched.
// Else clears the matrix when it actually redimensions it.

void Mat::dimClear (size_t nrows1, size_t ncols1)
{
if (nrows1 == 0 || ncols1 == 0)     // zero size?
    {
    if (m)
        {
        ASSERT(m->owner);           // only the owner can change the size (a view can't)
        gsl_matrix_free(m);
        m = NULL;
        }
    }
else if (!m)
    m = gsl_matrix_calloc(nrows1, ncols1);
else if (nrows() != nrows1 || ncols() != ncols1)
    {
    ASSERT(m->owner);
    gsl_matrix_free(m);
    m = gsl_matrix_calloc(nrows1, ncols1);
    }
}

//-----------------------------------------------------------------------------
// This preserves as much of the old data as possible.  If new matrix is
// bigger than or same size as the old matrix then all data will be preserved.
// Unused entries in the new matrix are cleared i.e. set to 0.

void Mat::dimKeep (size_t nrows1, size_t ncols1)
{
if (nrows1 == 0 || ncols1 == 0) // zero size?
    {
    if (m)
        {
        ASSERT(m->owner);   // only the owner can change the size (a view can't)
        gsl_matrix_free(m);
        m = NULL;
        }
    }
else if (!m)
    m = gsl_matrix_calloc(nrows1, ncols1);
else if (nrows() != nrows1 || ncols() != ncols1)
    {
    ASSERT(m->owner);
    gsl_matrix *p = gsl_matrix_calloc( nrows1, ncols1);

    // copy as much of the data as will fit in the new matrix

    size_t nMinrows = MIN(nrows1, nrows());
    size_t nMincols = MIN(ncols1, ncols());
    for (size_t i = 0; i < nMinrows; i++)
        for (size_t j = 0; j < nMincols; j++)
            gsl_matrix_set(p, i, j, gsl_matrix_get(m, i, j));
    gsl_matrix_free(m);
    m = p;                  // copy top structure only, not values, for speed
    }
}

//-----------------------------------------------------------------------------
// Redimension this to the same size as other and copy other into this
// If no redimensioning is needed this is about as fast as operator=

void Mat::assign (const Mat &other)
{
dim(other.nrows(), other.ncols());
copy(other);
}

//-----------------------------------------------------------------------------
// helper function

const Mat Mat::reshapeAux (const Mat &A, size_t nNewRows, size_t nNewCols,
            bool fRowMajor, size_t nStartRow, size_t nStartCol,
            size_t nrows1, size_t ncols1)
{
if (nrows1 == 0)
    nrows1 = A.nrows() - nStartRow;
if (ncols1 == 0)
    ncols1 = A.ncols() - nStartCol;

#if GSL_RANGE_CHECK
ASSERT(nStartRow >= 0 && nStartCol >= 0);

if (nStartRow + nrows1 > A.nrows() || nStartCol + ncols1 > A.ncols())
    Err("Mat::reshape rows %d+%d > %d or columns %d+%d > %d",
        nStartRow, nrows1, A.nrows(), nStartCol, ncols1, A.ncols());

if (nrows1 * ncols1 < nNewRows * nNewCols)
    Err("Mat::reshape old dimensions %dx%d < new dimensions %dx%d",
        nrows1, ncols1, nNewRows, nNewCols);
#endif

Mat m(nNewRows, nNewCols);
size_t i0 = 0, j0 = 0;
if (fRowMajor)
    {
    for (size_t j = nStartCol; j < nStartCol + ncols1; j++)
        for (size_t i = nStartRow; i < nStartRow + nrows1; i++)
            {
            m(i0, j0) = A(i,j);
            if (++i0 == nNewRows)
                {
                i0 = 0;
                j0++;
                if (j0 == nNewCols)
                    return m;
                }
            }
    }
else
    {
    for (size_t i = nStartRow; i < nStartRow + nrows1; i++)
        for (size_t j = nStartCol; j < nStartCol + ncols1; j++)
            {
            m(i0, j0) = A(i,j);
            if (++j0 == nNewCols)
                {
                j0 = 0;
                i0++;
                if (i0 == nNewRows)
                    return m;
                }
            }
    }
ASSERT(false);  // should never get here
return m;
}

//-----------------------------------------------------------------------------
// Return a submatrix of this, redimensioned to nNewRows and nNewCols.
// If you just want a submatrix then use submatrix().
//
// nStartRow nStartCol nrows1 ncols1 specify the submatrix in the original matrix.
//
// nNewRows and nNewCols specify the layout in the new matrix.
//
// If nrows1=0 it will be set to max possible, ditto for ncols1.
//
// Thus if just want to relayout the rows and cols of a matrix then set all
// of nStartRow nStartRow nStartCol nrows1 ncols1 to 0 and nNewRows and
// nNewCols to the new layout.
//
// If you want to access parts of a matrix without data copying, perhaps a
// view will suit you better -- see MatView in matview.hpp.  Using a view
// is usually more efficient.

Mat Mat::reshape (size_t nNewRows, size_t nNewCols, bool fRowMajor,
                size_t nStartRow, size_t nStartCol, size_t nrows1, size_t ncols1)
{
return reshapeAux(*this, nNewRows, nNewCols, fRowMajor,
                  nStartRow, nStartCol, nrows1, ncols1);
}

//-----------------------------------------------------------------------------
// Make me a submatrix of myself, redimensioned to nNewRows and nNewCols.
// See header comments in reshape().

Mat Mat::reshapeMe (size_t nNewRows, size_t nNewCols, bool fRowMajor,
                size_t nStartRow, size_t nStartCol, size_t nrows1, size_t ncols1)
{
Mat result(reshapeAux(*this, nNewRows, nNewCols,
        fRowMajor, nStartRow, nStartCol, nrows1, ncols1));

this->assign(result);
return *this;
}

//-----------------------------------------------------------------------------
void Mat::printAux (const char sMsg[], const char sFormat[],    // helper for print()
                    const char *sFile, FILE *pFile, size_t nMax, bool fPrintRowIndex) const
{
bool fCloseFile = false;

ASSERT(!(pFile && !sFile)); // must give a file name for error reporting if pFile is given

if (sFile && pFile == NULL)
    {
    pFile = fopen(sFile, "w");
    if (!pFile)
        Err("Can't open %s", sFile);
    fCloseFile = true;
    }
if (!sFile) // make sure we have something to print in user messages
    sFile = "file";

if (nMax == 0)
    nMax = MAX_MAT_DIM;
size_t nrows1 = MIN(nMax, nrows());
size_t ncols1 = MIN(nMax, ncols());

if (sMsg)
    if (pFile)
        fprintf(pFile, sMsg, nrows(), ncols());
    else
        lprintf(sMsg, nrows(), ncols());

for (size_t iRow = 0; iRow < nrows1; iRow++)
    {
    if (fPrintRowIndex)
        if (pFile)
            fprintf(pFile, "%3d: ", iRow);
        else
            lprintf("%3d: ", iRow);
    for (size_t iCol = 0; iCol < ncols1; iCol++)
        if (pFile)
            if (sFormat)
                fprintf(pFile, sFormat, getElem(iRow, iCol));
            else
                fprintf(pFile, MAT_PRINT_FORMAT, getElem(iRow, iCol));
        else
            if (sFormat)
                lprintf(sFormat, getElem(iRow, iCol));
            else
                lprintf(MAT_PRINT_FORMAT, getElem(iRow, iCol));
    if (iRow < nrows1)
        {
        if (pFile)
            fprintf(pFile, "\n");
        else
            lprintf("\n");
        }
    }
if (pFile)
    fflush(pFile);  // flush it out so it can be seen even if application crashes later
else
    fflush(stdout);

if (fCloseFile)
    fclose(pFile);
}

void Mat::print (const char sMsg[], const char sFormat[],
                    const char *sFile, FILE *pFile, int nMax, bool fPrintRowIndex) const
{
printAux(sMsg, sFormat, sFile, pFile, nMax, fPrintRowIndex);
}

void Mat::print (const char sMsg[], const char sFormat[],
                    const char *sFile, FILE *pFile, int nMax, bool fPrintRowIndex)
{
printAux(sMsg, sFormat, sFile, pFile, nMax, fPrintRowIndex);
}

//-----------------------------------------------------------------------------
// you should init nrows() and ncols() before calling this routine

Mat& Mat::operator= (const double Vals[])   // assign an array of doubles
{
size_t iVal = 0;

for (size_t iRow = 0; iRow < nrows(); iRow++)
        for (size_t iCol = 0; iCol < ncols(); iCol++)
            gsl_matrix_set(m, iRow, iCol, Vals[iVal++]);

return *this;
}

//-----------------------------------------------------------------------------
bool Mat::operator== (const Mat &other) const
{
if (nrows() != other.nrows() || ncols() != other.ncols())
  return false;

for (size_t iRow = 0; iRow < nrows(); iRow++)
    for (size_t iCol = 0; iCol < ncols(); iCol++)
        if (this->getElem(iRow, iCol) != other.getElem(iRow, iCol))
            return false;

return true;
}

//-----------------------------------------------------------------------------
Mat Mat::operator+ (const Mat &other) const
{
Mat result(*this);
gsl_matrix_add(result.m, other.m);
return result;
}

//-----------------------------------------------------------------------------
Mat Mat::operator+ (const double &d) const
{
Mat result(*this);
gsl_matrix_add_constant(result.m, d);
return result;
}

//-----------------------------------------------------------------------------
Mat operator+ (const double &d, const Mat &other)
{
Mat result(other);
gsl_matrix_add_constant(result.m, d);
return result;
}

//-----------------------------------------------------------------------------
Mat &Mat::operator+= (const double &d)
{
gsl_matrix_add_constant(m, d);
return *this;
}

//-----------------------------------------------------------------------------
Mat &Mat::operator+= (const Mat &other)
{
gsl_matrix_add(m, other.m);
return *this;
}

//-----------------------------------------------------------------------------
Mat Mat::operator- (const Mat &other) const
{
Mat result(*this);
gsl_matrix_sub(result.m, other.m);
return result;
}

//-----------------------------------------------------------------------------
Mat Mat::operator- (const double &d) const
{
Mat result(*this);
gsl_matrix_add_constant(result.m, -d);
return result;
}

//-----------------------------------------------------------------------------
Mat operator- (const double &d, const Mat &other)
{
Mat result(-1 * other);
gsl_matrix_add_constant(result.m, d);
return result;
}

//-----------------------------------------------------------------------------
Mat &Mat::operator-= (const double &d)
{
gsl_matrix_add_constant(m, -d);
return *this;
}

//-----------------------------------------------------------------------------
Mat &Mat::operator-= (const Mat &other)
{
gsl_matrix_sub(m, other.m);
return *this;
}

//-----------------------------------------------------------------------------
// Note: this isn't consistent with operator+ because we don't use
// matrix/gsl_matrix_mul here.  This is because for matrix multiply we have to
// redimension the output and therefore we use linalg/gsl_linalg_matmult.

Mat Mat::operator* (const Mat &other) const
{
Mat result(nrows(), other.ncols());
gsl_linalg_matmult(m, other.m, result.m);
return result;
}

//-----------------------------------------------------------------------------
Mat Mat::operator* (const double &d) const
{
Mat result(*this);
gsl_matrix_scale(result.m, d);
return result;
}

//-----------------------------------------------------------------------------
Mat operator* (const double &d, const Mat &other)
{
Mat result(other);
gsl_matrix_scale(result.m, d);
return result;
}

//-----------------------------------------------------------------------------
Mat &Mat::operator*= (const double &d)
{
gsl_matrix_scale(m, d);
return *this;
}

//-----------------------------------------------------------------------------
Mat &Mat::operator*= (const Mat &other)
{
*this = (*this) * other;
return *this;
}

//-----------------------------------------------------------------------------
Mat Mat::operator/(const double &d) const
{
Mat result(*this);

if (USE_EQ_DOUBLES)
    ASSERT(!fEqual(d, 0));
else
    ASSERT(d != 0);

gsl_matrix_scale(result.m, 1.0 / d);
return result;
}

//-----------------------------------------------------------------------------
Mat &Mat::operator/= (const double &d)
{
if (USE_EQ_DOUBLES)
    ASSERT(!fEqual(d, 0));
else
    ASSERT(d != 0.0);

gsl_matrix_scale(m, 1.0 / d);
return *this;
}

//-----------------------------------------------------------------------------
double Mat::absSum () const
{
double sum = 0;
for (size_t iRow = 0; iRow < nrows(); iRow++)
    for (size_t iCol = 0; iCol < ncols(); iCol++)
        sum += fabs(getElem(iRow, iCol));
return sum;
}

//-----------------------------------------------------------------------------
Mat Mat::mulElems (const Mat &other) const
{
Mat result(*this);
gsl_matrix_mul_elements(result.m, other.m);
return result;
}

//-----------------------------------------------------------------------------
Mat Mat::mulElems (const Mat &other)
{
Mat result(*this);
gsl_matrix_mul_elements(result.m, other.m);
return result;
}

//-----------------------------------------------------------------------------
Mat Mat::divElems (const Mat &other) const
{
Mat result(*this);
gsl_matrix_div_elements(result.m, other.m);
return result;
}

//-----------------------------------------------------------------------------
Mat Mat::divElems (const Mat &other)
{
Mat result(*this);
gsl_matrix_div_elements(result.m, other.m);
return result;
}

//-----------------------------------------------------------------------------
// Apply the given function to each element in matrix.
// See the definitions of tpMatFunc at top of mat.hpp to see meaning of parameters.
//
// Examples:
//
// A.map(sqrt) returns a mat where each elem is the root of the coresponding element in A.
//
// double times2(double Elem) { return 2 * x; }
// A.map(times2) returns a mat where each elem is twice the coresponding element in A.

Mat Mat::map (tpMatFunc pFunc) const
{
Mat result(*this);
for (unsigned iRow = 0; iRow < nrows(); iRow++)
    for (unsigned iCol = 0; iCol < ncols(); iCol++)
        gsl_matrix_set(result.m, iRow, iCol, pFunc(getElem(iRow, iCol)));

return result;
}

//-----------------------------------------------------------------------------
// utility function for squareElems

static inline double __cdecl square1 (double x)
{
return x * x;
}


//-----------------------------------------------------------------------------
bool fSameDim (const Mat &A, const Mat &B)
{
return A.nrows() == B.nrows() && A.ncols() == B.ncols();
}

//-----------------------------------------------------------------------------
void CheckReasonableDim (size_t nrows, size_t ncols, size_t nMin, const char *sMsg)
{
if ((size_t)nrows < nMin || nrows > MAX_MAT_DIM ||
        (size_t)ncols < nMin || ncols > MAX_MAT_DIM)
    Err(sMsg, nrows, ncols);
}

//-----------------------------------------------------------------------------
void CheckDim (const Mat &A, size_t nrows, size_t ncols, const char *sMsg)
{
if (A.nrows() != nrows || A.ncols() != ncols)
    Err("%s wrong size matrix %dx%d expected %dx%d",
        sMsg, A.nrows(), A.ncols(), nrows, ncols);
}

//-----------------------------------------------------------------------------
void CheckSameNbrRows (const Mat &A, const Mat &B, const char *sMsg)
{
if (A.nrows() != B.nrows())
    Err("%s matrices have different number of rows %dx%d %dx%d",
        sMsg, A.nrows(), A.ncols(), B.nrows(), B.ncols());
}

//-----------------------------------------------------------------------------
void CheckIsVector (const Mat &A, const char *sMsg)
{
if (!A.fVector())
    Err("%s %dx%d isn't a vector", sMsg, A.nrows(), A.ncols());
}

//-----------------------------------------------------------------------------
void CheckIsSquare (const Mat &A, const char *sMsg)
{
if (!A.fSquare())
    Err("%s matrix %dx%d isn't square", sMsg, A.nrows(), A.ncols());
}

//-----------------------------------------------------------------------------
void CheckSameDim (const Mat &A, const Mat &B, const char *sMsg)
{
if (!fSameDim(A, B))
    Err("%s matrices have different sizes %dx%d %dx%d",
        sMsg, A.nrows(), A.ncols(), B.nrows(), B.ncols());
}

//-----------------------------------------------------------------------------
double Mat::trace () const
{
double sum = 0;
#if GSL_RANGE_CHECK
CheckIsSquare(*this, "Mat trace");
#endif
for (unsigned iRow = 0; iRow < nrows(); iRow++)
    sum += getElem(iRow, iRow);
return sum;
}

//-----------------------------------------------------------------------------
// Dot product.
// This is an efficient implementation of:  mulElems(other)).sum().

double Mat::dot (const Mat &other) const
{
const size_t nrows1 = this->nrows();
const size_t ncols1 = this->ncols();
ASSERT(other.nrows() == nrows1 && other.ncols() == ncols1);

double result = 0;
const size_t thisTda   = this->m->tda;
const size_t otherTda  = other.m->tda;
const double *thisData = this->m->data;
const double *otherData = other.m->data;

// for speed we check for special cases

if (nrows1 == 1)                        // column vector?
    for (size_t j = 0; j < ncols1; j++)
        result += thisData[j] * otherData[j];

else if (ncols1 == 1)                   // row vector?
    for (size_t i = 0; i < nrows1; i++)
        result += thisData[i * thisTda] * otherData[i * otherTda];

else                                    // not a vector, must be a matrix
    for (size_t i = 0; i < nrows1; i++)
        {
        const size_t iThis  = i * thisTda;
        const size_t iOther = i * otherTda;
        for (size_t j = 0; j < ncols1; j++)
            result += thisData[iThis + j] * otherData[iOther + j];
        }

return result;
}

//-----------------------------------------------------------------------------
Mat Mat::diagMe (double d)
{
unsigned n;
if (nrows() < ncols())
    n = nrows();
else
    n = ncols();
for (size_t i = 0; i < n; i++)
    gsl_matrix_set(m, i, i, d);
return *this;
}

//-----------------------------------------------------------------------------
// true if a row is all zeroes, or all equal to Val

bool Mat::fARowIsZero (double Val) const
{
for (size_t iRow = 0; iRow < nrows(); iRow++)
    {
    bool fAllZero = true;
    for (size_t iCol = 0; iCol < ncols(); iCol++)
        if (getElem(iRow, iCol) != Val)
            {
            fAllZero = false;
            break;
            }
    if (fAllZero)
        return true;
    }
return false;
}

//-----------------------------------------------------------------------------
bool Mat::fAllZeroes (double Val) const // true if all values zero, or equal to Val
{
bool fAllZero = true;
for (unsigned iRow = 0; iRow < nrows(); iRow++)
    for (unsigned iCol = 0; iCol < ncols(); iCol++)
        if (getElem(iRow, iCol) != Val)
            {
            fAllZero = false;
            break;
            }

return fAllZero;
}

//-----------------------------------------------------------------------------
double Mat::det (int sign) const
{
Mat a = *this;
a = a.LU_decomp();
return gsl_linalg_LU_det(a.m, sign);
}

//-----------------------------------------------------------------------------
double Mat::lndet () const
{
Mat a = *this;
a = a.LU_decomp();
return gsl_linalg_LU_lndet(a.m);
}

//-----------------------------------------------------------------------------
// If pFile is NULL then it is opened.  If pFile is already open then we read
// from it (and sFile isn't used except for user messages).
//
// If we try to read a matrix and can't find the matrix header '{' then we
// invoke Err().
//
// We copy the matrix tag string into sTag, if sTag is not null.
// The caller must ensure that there is enough space in sTag.
// A tag string is an optional string enclosed in quotes before
// the matrix first delimiter '{'.
// It can be used for storing filenames and other info for the matrix.
// The leading and closing quotes are removed before copying into sTag.
// If there is no tag string, sTag[0] is set to 0.
//
// File format: see Mat::write

void Mat::read (const char *sFile, FILE *pFile, char *sTag)
{
char sTag1[256]; sTag1[0] = 0;  // mat prefix we read from file, possibly copied into sTag
if (sTag)
    sTag[0] = 0;

bool fCloseFile = false;

ASSERT(!(pFile && !sFile)); // need a file name for error reporting if pFile is given

if (pFile == NULL)
    {
    ASSERT(sFile);
    pFile = fopen(sFile, "r");
    if (!pFile)
        Err("Can't open %s", sFile);
    fCloseFile = true;
    }
size_t nrows, ncols;

if (!sFile) // make sure we have something to print for user messages
    sFile = "file";

// search for first delimiter '{', skipping optional comments prefixed by # or "
char c;
do {
    if (1 != ::fread(&c, 1, 1, pFile))
        Err("Can't read matrix header from %s (no '{')", sFile);
    if (c == '#')                       // comment?
        {
        while (c != '\n')               // if so, then skip it
            if (1 != ::fread(&c, 1, 1, pFile))
                Err("Can't read matrix header from %s (no '{') ", sFile);
        }
    else if (c == '"')                  // embedded string?
        {
        if (!fgets(sTag1, 256-1, pFile))
            Err("Can't read matrix header from %s (no '{')", sFile);
        if (sTag)
            {
            char *pDest = strchr(sTag1, '"');
                                // there may or may not be a terminating ", be lenient
            int iLen = strlen(sTag1)-1;
            if (pDest)
                {
                iLen--;
                *pDest = 0;
                }
            else
                sTag1[iLen] = 0;
            strcpy(sTag, sTag1);
            }
        }
    }
while (c != '{');

if (2 != ::fscanf(pFile, "%u %u\n", &nrows, &ncols))
    Err("Can't read matrix size from %s", sFile);

char sErr[256];
sprintf(sErr, "Implausible number of rows %d or cols %d in %s (file corrupt?) ",
     nrows, ncols, sFile);
CheckReasonableDim(nrows, ncols, 0, sErr);

free();
if (nrows && ncols) // only read if there is some data
    {
    m = gsl_matrix_alloc(nrows, ncols);
    if (0 != gsl_matrix_fscanf(pFile, m))
        Err("Can't read matrix from %s (file truncated?) ", sFile);
    }
c = ' ';
while (c == ' ' || c == '\t' || c == '\n' || c == '\r') // skip white space
    if (1 != ::fread(&c, 1, 1, pFile))
        Err("Can't read matrix from %s (reached EOF before finding '}' footer) ", sFile);

if (c != '}')
    Err("%s footer not '}'", sFile);

if (fCloseFile)
    fclose(pFile);
}

//-----------------------------------------------------------------------------
Mat Mat::t () const             // transpose of matrix
{
Mat result(ncols(), nrows());

for (size_t i = 0; i < nrows(); i++)
    for (size_t j = 0; j < ncols(); j++)
        gsl_matrix_set(result.m, j, i, gsl_matrix_get(m, i, j));

return result;
}

//-----------------------------------------------------------------------------
Mat Mat::inverse () const
{
gsl_permutation *pPerm = gsl_permutation_alloc(nrows());
Mat A = *this;
A = A.LU_decomp(pPerm);
Mat AInverse(nrows(), ncols());
int iRet = gsl_linalg_LU_invert(A.m, pPerm, AInverse.m);
ASSERT(iRet == 0);
gsl_permutation_free(pPerm);
return AInverse;
}

//-----------------------------------------------------------------------------
Mat Mat::normalizeMe ()
{
double Len = this->len();

// We want to be able to normalize zero length vectors without worrying about
// about div by zero or checks exterior to this function. So we do a check here.

if (!fEqual(Len, 0))
    gsl_matrix_scale(this->m, 1.0 / Len);

return *this;
}

Mat Mat::normalize () const
{
Mat result(*this);
double Len = this->len();
if (!fEqual(Len, 0))
    gsl_matrix_scale(result.m, 1.0 / Len);
return result;
}

//-----------------------------------------------------------------------------
double Mat::sum () const
{
double sum = 0;
for (size_t iRow = 0; iRow < nrows(); iRow++)
    for (size_t iCol = 0; iCol < ncols(); iCol++)
        sum += getElem(iRow, iCol);
return sum;
}

//-----------------------------------------------------------------------------
Mat Mat::LU_decomp (gsl_permutation *pPerm, int *pSign) const
{
bool fRetPerm = (pPerm != NULL);
if (!pPerm)
    pPerm = gsl_permutation_alloc(nrows());
int Sign;
Mat result = *this;
int iRet = gsl_linalg_LU_decomp(result.m, pPerm, &Sign);

// check if matrix is singular i.e. has a zero pivot
for (int i = result.nrows()-1; i >= 0; i--)    // smallest last so look there 1st
    ASSERT(!fEqual(result(i,i), FLT_EPSILON)); // use FLT_EPSILON as a convenient small nbr

ASSERT(iRet == 0);

if (!fRetPerm)
    gsl_permutation_free(pPerm);
if (pSign)
    *pSign = Sign;
return result;
}

//-----------------------------------------------------------------------------
// Solves Ax=b by LU decomposition.  Returns col vec x and LU factorization in LU.
// A doesn't get changed.
// If A is singular this will fail (it doesn't do least squares or anything like that)

Vec SolveWithLU (const Mat &A, Vec &b)
{
#if GSL_RANGE_CHECK
CheckIsVector(b, "SolveWithLU b");
if (A.nrows() != b.nelems())
    Err("SolveWithLU b %dx%d doesn't conform to A %dx%d",
        b.nrows(), b.ncols(), A.nrows(), A.ncols());
#endif

int Sign, iRet;
size_t nrows = A.nrows();

gsl_permutation *pPerm = gsl_permutation_alloc(nrows);
gsl_vector      *xVec  = gsl_vector_alloc(nrows);
gsl_vector_view  bVec  = gsl_vector_view_array(b.m->data, nrows);
gsl_matrix      *pLU   = gsl_matrix_alloc(nrows, A.ncols());

gsl_matrix_memcpy(pLU, A.m);    // use copy not view because decomp destroys A

if (0 != (iRet = gsl_linalg_LU_decomp(pLU, pPerm, &Sign)))
    Err("SolveWithLU gsl_linalg_LU_decomp returned %d", iRet);

if (0 != (iRet = gsl_linalg_LU_solve(pLU, pPerm, &bVec.vector, xVec)))
    Err("SolveWithLU gsl_linalg_LU_solve returned %d", iRet);

Vec x(xVec->data, nrows);   // convert gsl_vector to Vec
gsl_permutation_free(pPerm);
gsl_vector_free(xVec);
gsl_matrix_free(pLU);

return x;
}

//-----------------------------------------------------------------------------
// Multiply a 3x3 matrix by a 1x2 vector. Used for homogenous transforms.
//
// TODO Could be made more efficient by using arrays internally instead of matrices.

Vec Mat::mat33TimesVec2 (const Vec &v2) const   // returns row Vec(2)
{
#if GSL_RANGE_CHECK
CheckDim(v2,    1, 2, "mat33TimesVec2 vector");
CheckDim(*this, 3, 3, "mat33TimesVec2 transform matrix");
#endif

static Vec Vec3(3, ROWVEC); // static for efficiency
Vec3(VX) = getElem(0, VX) * v2(VX)  +  getElem(0, VY) * v2(VY)  +  getElem(0, VZ);
Vec3(VY) = getElem(1, VX) * v2(VX)  +  getElem(1, VY) * v2(VY)  +  getElem(1, VZ);
Vec3(VZ) = getElem(2, VX) * v2(VX)  +  getElem(2, VY) * v2(VY)  +  getElem(2, VZ);

ASSERT(!fEqual(Vec3(0, VZ), 0));    // check for div by zero

static Vec Vec2(2, ROWVEC);
Vec2(VX) = Vec3(0, VX) / Vec3(0, VZ);
Vec2(VY) = Vec3(0, VY) / Vec3(0, VZ);

return Vec2;
}

//-----------------------------------------------------------------------------
MatView Mat::view (size_t iStartRow, size_t iStartCol,
                    size_t nrows1, size_t ncols1, size_t Tda)
{
MatView View(*this, iStartRow, iStartCol, nrows1, ncols1, Tda);
return View;
}

//-----------------------------------------------------------------------------
const MatView Mat::view (size_t iStartRow, size_t iStartCol,
                        size_t nrows1, size_t ncols1, size_t Tda) const
{
MatView View(*this, iStartRow, iStartCol, nrows1, ncols1, Tda);
return View;
}

//-----------------------------------------------------------------------------
VecView Mat::row (size_t iRow)
{
return VecView(*this, iRow, 0, 1, this->ncols());
}

//-----------------------------------------------------------------------------
const VecView Mat::row (size_t iRow) const
{
return VecView(*this, iRow, 0, 1, this->ncols());
}

//-----------------------------------------------------------------------------
VecView Mat::rowAsCol (size_t iRow)
{
return VecView(*this, iRow, 0, this->ncols(), 1, 1);
}

//-----------------------------------------------------------------------------
const VecView Mat::rowAsCol (size_t iRow) const
{
return VecView(*this, iRow, 0, this->ncols(), 1, 1);
}

//-----------------------------------------------------------------------------
VecView Mat::col (size_t iCol)
{
return VecView(*this, 0, iCol, this->nrows(), 1);
}

//-----------------------------------------------------------------------------
const VecView Mat::col (size_t iCol) const
{
return VecView(*this, 0, iCol, this->nrows(), 1);
}

//-----------------------------------------------------------------------------
VecView Mat::viewDiagAsCol ()
{
return VecView(*this, 0, 0, this->nrows(), 1, ncols()+1);
}

//-----------------------------------------------------------------------------
const VecView Mat::viewDiagAsCol () const
{
return VecView(*this, 0, 0, this->nrows(), 1, ncols()+1);
}

//-----------------------------------------------------------------------------
VecView Mat::viewAsRow ()
{
return VecView(*this, 0, 0, 1, this->nrows() * this->ncols());
}

//-----------------------------------------------------------------------------
const VecView Mat::viewAsRow () const
{
return VecView(*this, 0, 0, 1, this->nrows() * this->ncols());
}

//-----------------------------------------------------------------------------
VecView Mat::viewAsCol ()
{
return VecView(*this, 0, 0, this->nrows() * this->ncols(), 1, 1);
}

//-----------------------------------------------------------------------------
const VecView Mat::viewAsCol () const
{
return VecView(*this, 0, 0, this->nrows() * this->ncols(), 1, 1);
}

//-----------------------------------------------------------------------------
MatView Mat::viewAsSquare ()
{
size_t ncols = size_t(sqrt(this->nelems()));
// check that can square properly (assertion is not essential but protects the user)
ASSERT(ncols * ncols == this->nelems());
return MatView(*this, 0, 0, ncols, ncols, ncols);
}

//-----------------------------------------------------------------------------
const MatView Mat::viewAsSquare () const
{
size_t ncols = size_t(sqrt(this->nelems()));
// check that can square properly (assertion is not essential but protects the user)
ASSERT(ncols * ncols == this->nelems());
return MatView(*this, 0, 0, ncols, ncols, ncols);
}

//-----------------------------------------------------------------------------
// returns a row vector whose elements are the sums of the columns

Vec Mat::colSums () const
{
Vec Cols(ncols(), ROWVEC);  // initialized to 0
int iRow = nrows();
while (iRow--)
    Cols += row(iRow);
return Cols;
}

//-----------------------------------------------------------------------------
// Set elements in the given range to Val.
// See also fill(const double &d) in mat.hpp.

void Mat::fill (double Val, size_t iStartRow, size_t iStartCol,
                size_t nEndRow, size_t nEndCol)
{
for (size_t iRow = iStartRow; iRow < nEndRow; iRow++)
    for (size_t iCol = iStartCol; iCol < nEndCol; iCol++)
        gsl_matrix_set(m, iRow, iCol, Val);
}

//-----------------------------------------------------------------------------
// Type convert a one element matrix to a double.
// It's probably possible to do this type conversion automatically but making
// the programmer do it explicitly helps catch errors.

double OneElemToDouble (const Mat &A)
{
DASSERT(A.nrows() == 1 && A.ncols() == 1);
return A(0, 0);
}

//-----------------------------------------------------------------------------
// This returns x * A * x'.
//
// x is a vector.
// A is assumed to be a symmetric matrix
// (but only the upper right triangle of A is actually used).
//
// This function is equivalent to OneElemToDouble(x * A * x.t()), but is
// optimized for speed and is much faster.  It's faster because we use the
// fact that A is symmetric to roughly halve the number of operations.  We
// also bypass the overhead of normal matrix routines, which can be slow.

double xAx (const Vec &x, const Mat &A)
{
// copy x into here because it's faster accessing an array than a Vec
double xTemp[MAX_TEMP_ARRAY_LEN];

const unsigned nelems = x.nelems();

ASSERT(nelems < unsigned(MAX_TEMP_ARRAY_LEN));
ASSERT(A.nrows() == nelems);
ASSERT(A.ncols() == nelems);
ASSERT(x.nrows() == 1);

double DiagResult = 0, Result = 0;

// init xTemp and sum diag elements

size_t i;
for (i = 0; i < nelems; i++)
    {
    xTemp[i] = x(i);
    DiagResult += A(i, i) * xTemp[i] * xTemp[i];
    }
// sum upper right triangle elems

for (i = 0; i < nelems; i++)
    {
    double xi = xTemp[i];
    for (size_t j = i+1; j < nelems; j++)
        Result += A(i, j) * xi * xTemp[j];
    }
Result *= 2;    // incorporate lower left triangle elements

return DiagResult + Result;
}

}   // end namespace GslMat
