// $find.cpp 1.5 milbo$
// Derived from code by Henry Rowley http://vasc.ri.cmu.edu/NNFaceDetector.
//
// The original files had the following header:
// By Henry A. Rowley (har@cs.cmu.edu, http://www.cs.cmu.edu/~har)
// Developed at Carnegie Mellon University for the face detection project.
// Code may be used, but please provide appropriate acknowledgements, and let
// har@cs.cmu.edu how you are using it.  Thanks!

#include "all.hpp"
#include "list.cpp"

static const char *NET_0 = "umec";      // neural net names
static const char *NET_1 = "face17c";
static const char *NET_2 = "face18c";
static const char *NET_3 = "eye";

Image gFaceMask;
Image gEyeMask;

//-----------------------------------------------------------------------------
// Save detection information in callbackData, which is a list of detections.
// The parameters match the typedef tCallbackFunc in search.hpp.

static void
SaveDetections (ClientData callbackData,
     Image *pImg, int x, int y, int width, int height,
     int iLev, double scale, double output)
{
List<Detection> *detectionList=(List<Detection> *)callbackData;
detectionList->addLast(new Detection(x, y, iLev, output));
}

//-----------------------------------------------------------------------------
// sFile must be of the form facemask20x20.pgm i.e. name followed by digits
// since we do a basic parse of the file name to extract the mask name

static void
LoadMask (Image &Mask,          // out
          const char sFile[])   // in
{
char sMask[FLEN];

// extract training mask e.g. "facemask20x20.pgm" becomes "facemask"

int i;
for (i = 0; sFile[i] && sFile[i] >= 'A' && sFile[i] <= 'z'; i++)
    sMask[i] = sFile[i];
sMask[i] = 0;
_strlwr(sMask);

sLoadImage(Mask, sFile);
}

//-----------------------------------------------------------------------------
// Initialize the pNetworkNames used by the Track_FindAllFaces function.
// Also init the globals gFaceMask and gEyeMask.

static void
Init (const char *sDataDir)
{
static bool fInitialized = false;
if (fInitialized)
    return;

// following must be static for InitNetwork
static char s0[SLEN], s1[SLEN], s2[SLEN], s3[SLEN];
sprintf(s0, "%s/%s", sDataDir, NET_0);
sprintf(s1, "%s/%s", sDataDir, NET_1);
sprintf(s2, "%s/%s", sDataDir, NET_2);
sprintf(s3, "%s/%s", sDataDir, NET_3);
const char *pNetworkNames[]={s0, s1, s2, s3};

AllocateNetworks(4);
InitNetwork(0, 1, pNetworkNames);       // network 0: umec
InitNetwork(1, 1, pNetworkNames+1);     // network 1: face17c
InitNetwork(2, 1, pNetworkNames+2);     // network 1: face18c
InitNetwork(3, 1, pNetworkNames+3);     // network 2: eye

char sFace[SLEN];                       // load face mask
sprintf(sFace, "%s/facemask20x20.pgm", sDataDir);
LoadMask(gFaceMask, sFace);

char sMask[SLEN];                       // load eye mask
sprintf(sMask, "%s/eyemask25x15.pgm", sDataDir);
LoadMask(gEyeMask, sMask);

// a consistency check -- prevents overflow in FindEyes()

if (gNetList[3]->nInputs - 1 != gEyeMask.width * gEyeMask.height)
    Err("Network \"%s\" number of inputs %d does not match %s size %d",
        pNetworkNames[3], gNetList[3]->nInputs-1,
        sFace, gEyeMask.width * gEyeMask.height);

fInitialized = true;
}

//-----------------------------------------------------------------------------
// Resample an image, according to the following rule: if any of the
// pixels in the source image which correspond to a pixel in the
// destination image is zero, then set the destination pixel to zero.
// Otherwise, set the pixel to 255.
// The idea is that 0 means a pixel location should be scanned, while
// 255 means that a face has already been found and need not be checked
// again.  Destination can be the same as the source.
// Scale < 1 makes the image smaller.
// The scaling factor is determined by the level in the image pyramid,
// and the size of the destination image is controlled by nNewWith and nNewHeight.

static void
ResampleMask (Image &Mask,                                  // out
              const Image &Src,                             // in
              int nNewWidth, int nNewHeight, double Scale)  // in
{
Image TempImg(nNewWidth, nNewHeight);
byte  *buf = TempImg.buf;

FillImage(TempImg, 255);

for (int iy = 0; iy < nNewHeight; iy++)     // scan over lower resolution image
    for (int ix = 0; ix < nNewWidth; ix++)
        {
        int ix1 = int(0.5 + ix / Scale);
        if (ix1 >= Src.width)
            ix1 = Src.width-1;
        int iy1 = int(0.5 + iy / Scale);
        if (iy1 >= Src.height)
            iy1 = Src.height-1;
        int ix2 = int(0.5 + (ix + 1) / Scale);
        if (ix2 >= Src.width)
            ix2 = Src.width-1;
        int iy2 = int(0.5 + (iy + 1) / Scale);
        if (iy2 >= Src.height)
            iy2 = Src.height-1;
        int i, j;
        for (j = iy1; j <= iy2; j++)    // scan over corresponding pixels in hi res
            for (i = ix1; i <= ix2; i++)
                if (!Src(i,j))
                    {                   // if any pixel is zero, make new pixel zero
                    *buf = 0;
                    goto next;
                    }
        next:
            buf++;
        }
Mask = TempImg;
}


//-----------------------------------------------------------------------------
// Figure out the scale at which to run the eye detector.
// The "7" was determined empirically.

static int GetEyeScale (int FaceScale)
{
    int EyeScale = FaceScale - 7;
    if (EyeScale >= - 3 && EyeScale < 0)
        EyeScale = 0;
    return EyeScale;
}

//-----------------------------------------------------------------------------
// Search for the eyes in the face and return the two eye positions.
// Right and left are wrt. the viewer.

static void FindEyes (
    int *pxLeftEye, int *pyLeftEye,     // out: position, or -1 if can't find left eye
    int *pxRightEye, int *pyRightEye,   // out: ditto for right eye
    const Image Img,                    // in
    const Image &EyeMask,               // in
    int xFace, int yFace,               // in: position of face in image
    int EyeScale,                       // in:
    int FaceScale,                      // in: scale in pyramid where image is
    int iNet)                           // in: neural net index in gNetList
{
*pxLeftEye = *pyLeftEye = *pxRightEye = *pyRightEye = -1;   // assume won't find eyes

int EyeMaskSize  = EyeMask.width * EyeMask.height;
int EyeMaskWidth = EyeMask.width;
int EyeMaskHeight = EyeMask.height;

double xLeft = 0, yLeft = 0, nLeft = 0, xRight = 0, yRight = 0, nRight = 0;

int *pTmpImage = new int[EyeMaskSize];  // window for eye detector

// possible upper-left X positions for the left eye

int startxLeft = int(floor(0.5 + xFace * POW12(FaceScale - EyeScale)) - EyeMaskWidth / 2);
int endxLeft = int(floor(0.5 + (xFace + 10) * POW12(FaceScale - EyeScale)) - EyeMaskWidth / 2);
if (startxLeft < 0)
    startxLeft = 0;
if (endxLeft >= Img.width - EyeMaskWidth + 1)
    endxLeft = Img.width - EyeMaskWidth + 1;

// possible upper-left X positions for the right eye

int startxRight = int(floor(0.5 + (xFace + 10) * POW12(FaceScale - EyeScale)) - EyeMaskWidth / 2);
int endxRight = int(floor(0.5 + (xFace + 20) * POW12(FaceScale - EyeScale)) - EyeMaskWidth / 2);
if (startxRight < 0)
    startxRight = 0;
if (endxRight >= Img.width - EyeMaskWidth + 1)
    endxRight = Img.width - EyeMaskWidth + 1;

// possible upper-left Y positions for the eyes

int starty = int(floor(0.5 + yFace * POW12(FaceScale - EyeScale)) - EyeMaskHeight / 2);
int endy = int(floor(0.5 + (yFace + 10) * POW12(FaceScale - EyeScale)) - EyeMaskHeight);
if (starty < 0)
    starty = 0;
if (endy >= Img.height - EyeMaskHeight + 1)
    endy = Img.height - EyeMaskHeight + 1;

for (int y = starty; y < endy; y++)
    {
    int i, j, x;

    // Look for right eye on this scan line.  We mirror it so it looks to the
    // net like a right eye because we trained the net on a left eye.

    for (x = startxRight; x < endxRight; x++)
        {
        int iPixel = 0;
        int Hist[256];
        memset(Hist, 0, sizeof(Hist));

        // copy the window into pTmpImage (using mirror image), and compute
        // the histogram over the entire window

        for (j = 0; j < EyeMaskHeight; j++)
            for (i = EyeMaskWidth - 1; i >= 0; i--) // note: mirror here
                {
                int Pixel = Img(i+x, j+y);
                pTmpImage[iPixel++] = Pixel;
                Hist[Pixel]++;
                }
        // compute cumulative histogram

        int CumHist[256];
        int Sum = 0;
        for (i = 0; i < 256; i++)
            {
            CumHist[i] = Sum;
            Sum += Hist[i];
            }
        int Total = Sum;
        for (i = 255; i >= 0; i--)
            {
            CumHist[i] += Sum;
            Sum -= Hist[i];
            }
        // apply the histogram equalization, and write window to network inputs

        const double Scale = 1.0 / Total;
        ForwardUnit *pUnit = &(gNetList[iNet]->pUnitList[1]);
        for (i = 0; i < EyeMaskSize; i++)
            (pUnit++)->activation = (_FLOAT)(CumHist[pTmpImage[i]] * Scale - 1.0);

        // if the network responds positively, add the detection to centroid

        const double Output = ForwardPass(gNetList[iNet]);
        if (Output > 0)
            {
            nRight += Output;
            xRight += Output * (x + EyeMaskWidth / 2)  * POW12(EyeScale);
            yRight += Output * (y + EyeMaskHeight / 2) * POW12(EyeScale);
            }
        }
    // look for left eye in this scan line

    for (x = startxLeft; x < endxLeft; x++)
        {
        int iPixel = 0;
        int Hist[256];
        memset(Hist, 0, sizeof(Hist));

        // copy the window into pTmpImage,  and compute
        // the histogram over the entire window

        for (j = 0; j < EyeMaskHeight; j++)
            for (i = 0; i < EyeMaskWidth; i++)
                {
                int Pixel = Img(i+x, j+y);
                pTmpImage[iPixel++] = Pixel;
                Hist[Pixel]++;
                }

        // compute cumulative histogram

        int CumHist[256];
        int Sum = 0;
        for (i = 0; i < 256; i++)
            {
            CumHist[i] = Sum;
            Sum += Hist[i];
            }
        int Total = Sum;
        for (i = 255; i >= 0; i--)
            {
            CumHist[i] += Sum;
            Sum -= Hist[i];
            }
        // apply the histogram equalization, and write window to network inputs

        const double Scale = 1.0 / Total;
        ForwardUnit *pUnit = &(gNetList[iNet]->pUnitList[1]);
        for (i = 0; i < EyeMaskSize; i++)
            (pUnit++)->activation = (_FLOAT)(CumHist[pTmpImage[i]] * Scale - 1.0);

        // if the network responds positively, add the detection to centroid

        const double Output = ForwardPass(gNetList[iNet]);
        if (Output > 0)
            {
            nLeft += Output;
            xLeft += Output * (x + EyeMaskWidth  / 2) * POW12(EyeScale);
            yLeft += Output * (y + EyeMaskHeight / 2) * POW12(EyeScale);
            }
        }
    }
// if the left eye was detected at least once, return centroid

if (nLeft > 0)
    {
    *pxLeftEye = int(floor(0.5 + xLeft / nLeft));
    *pyLeftEye = int(floor(0.5 + yLeft / nLeft));
    }
// if the right eye was detected at least once, return centroid

if (nRight > 0)
    {
    *pxRightEye = int(floor(0.5 + xRight / nRight));
    *pyRightEye = int(floor(0.5 + yRight / nRight));
    }
delete[] pTmpImage;
}

//-----------------------------------------------------------------------------
// Locate all faces and eyes in an image.
//
// Returns the number of faces found.
//
// If faces were found, *ppFaceLocations will point to an
// array of tFaceEyeLocation's.  Caller must free pFaceLocations.
// If no faces were found, *ppFaceLocations will be set to NULL.
//
// pFaceLocations will also contain the locations of the eyes (see
// FindEyes for details).
//
// Overview of this func:
// This first does a quick search scanning 10 pixels at a time using umec.net
// It then verifies results in this way: if both face17c.net and face18c.net
// agree on a detection then the centroid is computed.
// Finally the eyes are located using eye.net.

int
Track_FindAllFaces (tFaceEyeLocation **ppFaceLocations,     // out
                    const Image &Img, const char *sDataDir) // in
{
Init(sDataDir);                      // init gFaceMask etc.

int nLevs = 0;                       // figure out number of levels in pyramid
int w = Img.width, h = Img.height;   // (this assumes image has a reasonable aspect ratio)
while (w >= 20 && h >= 20)
    {
    w = int(w / 1.2);
    h = int(h / 1.2);
    nLevs++;
    }
Image *ImagePyramid = new Image[nLevs];    // create pyramid: an array of images
ImagePyramid[0] = Img;
int iLev;
for (iLev = 1; iLev < nLevs; iLev++)
    ReduceSizeBy1_2(ImagePyramid[iLev], ImagePyramid[iLev-1]);

Image MainMask(Img.width, Img.height, true);
                        // init to all zeroes: will later set 0=scan, 255=face found

Image LevMask;          // mask for current level

List<Detection> AllDetectionsList[1];   // where we store detections
for (iLev = nLevs-1; iLev >= 0; iLev--)
    {
    // get search mask at right level -- this scales MainMask into LevMask

    ResampleMask(LevMask, MainMask,
        ImagePyramid[iLev].width, ImagePyramid[iLev].height, POW12(-iLev));

    // Find candidates for this level in the pyramid.
    // We search the given image using the 30x30 candidate detector.
    // LevMask is used to indicate which portions of the image not to
    // search because a face has already been found there.
    // When a face is found, the callback func SaveDetections
    // is invoked which basically saves the face in DetectionsList.
    // Note that the face mask is not used at this point.
    // iNet is hardcoded to 0 in SearchUme so we use umec.net (inited in Init).

    List<Detection> DetectionsList;
    SearchUme(ImagePyramid[iLev], &LevMask, iLev,
                SaveDetections, (ClientData)(&DetectionsList));

    for (Detection* detect = DetectionsList.first; detect != NULL; detect = detect->next)
        {
        // Verify candidates. It works like this:
        // Given a candidate detection location and the image pyramid,
        // search the image around that candidate to try and find out
        // whether there is really a face there.
        // iNet=1 so we use nets 1 and 2 here (initialized
        // in Init to face17c.net and face18c.net).

        int newX, newY, newS;

        if (FindNewLocation(nLevs, ImagePyramid, gFaceMask,
                detect->x, detect->y, detect->scale,
                6, 6, 1, 2, 1, &newX, &newY, &newS)) // 6 dx, 6 dy, 1 ds, 2 step, 1 iNet
            {
            // got a verified candidate -- save it

            SaveDetections((ClientData)AllDetectionsList, &ImagePyramid[newS],
                    newX, newY, 20, 20, newS, POW12(newS), 1.0);

            // put detection in main search mask, to save computation

            FillRectangle(MainMask, (int)(POW12(newS)*newX + 0.5),
                    (int)(POW12(newS) * newY + 0.5),
                    (int)(POW12(newS) * (newX+20) + 0.5),
                    (int)(POW12(newS) * (newY+20) + 0.5), 255);
            }
        }
    }
// remove overlapping detections (a few may slip through)

List<Detection> ResultsList;
FuzzyVote2(Img.width, Img.height, 1, AllDetectionsList, SaveDetections,
    (ClientData)(&ResultsList),
    2, 1, true, true, false, // 2 spread, 1 search, collapse, overlap, fFilterOdd
    gFaceMask);

if (ResultsList.length == 0)
    *ppFaceLocations = NULL;
else
    {
    *ppFaceLocations =
        (tFaceEyeLocation*)malloc(sizeof(tFaceEyeLocation) * ResultsList.length);

    // locate the eyes in each detected face

    int iLocs = 0;
    for (Detection *d = ResultsList.first; d != NULL; d = d->next)
        {
        (*ppFaceLocations)[iLocs].xRightEye = -1; // assume won't find eyes
        (*ppFaceLocations)[iLocs].yRightEye = -1;
        (*ppFaceLocations)[iLocs].xLeftEye = -1;
        (*ppFaceLocations)[iLocs].yLeftEye = -1;

        int EyeScale = GetEyeScale(d->scale);
        if (EyeScale >= 0)  // face is big enough?
            FindEyes(
                &((*ppFaceLocations)[iLocs].xRightEye),
                &((*ppFaceLocations)[iLocs].yRightEye),
                &((*ppFaceLocations)[iLocs].xLeftEye),
                &((*ppFaceLocations)[iLocs].yLeftEye),
                ImagePyramid[EyeScale], gEyeMask,
                d->x, d->y, EyeScale, d->scale, 3);  // 3 is iNet i.e. eye.net

        // map face location to original resolution
        // (eye locations are already mapped correctly)

        double ImageScale = POW12(d->scale);
        (*ppFaceLocations)[iLocs].x1 = (int)floor(d->x * ImageScale + 0.5);
        (*ppFaceLocations)[iLocs].y1 = (int)floor(d->y * ImageScale + 0.5);
        (*ppFaceLocations)[iLocs].x2 = (int)floor((d->x + 20) * ImageScale + 0.5);
        (*ppFaceLocations)[iLocs].y2 = (int)floor((d->y + 20) * ImageScale + 0.5);

        iLocs++;
        }
    }
for (iLev = 1; iLev < nLevs; iLev++)
    ImagePyramid[iLev].free();

delete[] ImagePyramid;

return ResultsList.length;
}

//-----------------------------------------------------------------------------
#if MASM    // needed only by research code
void
FindEyesGivenFace (int *pxRightEye, int *pyRightEye,    // out: -1 if not found
        int *pxLeftEye, int *pyLeftEye,                 // out: -1 if not found
        const SHAPE DetShape,                           // in: posn of face
        const Image &Img, const char *sDataDir)         // in
{
Init(sDataDir);     // init gFaceMask etc.

*pxLeftEye = -1;    // assume we won't find the eyes
*pyLeftEye = -1;
*pxRightEye = -1;
*pyRightEye = -1;

double width = DetShape(DETECTOR_BotRight, VX) - DetShape(DETECTOR_TopLeft, VX);

int FaceScale = iRound(log(width / 20.0) / log(1.2));
int EyeScale = GetEyeScale(FaceScale);
if (EyeScale >= 0)  // face is big enough?
    {
    double ImageScale = POW12(FaceScale);

    int x = iRound((Img.width/2  + DetShape(DETECTOR_TopLeft, VX)) / ImageScale);

    int y = iRound((Img.height/2 - DetShape(DETECTOR_TopLeft, VY)) / ImageScale);

    Image ScaledImg(Img);
    ReduceImage(ScaledImg, POW12(EyeScale), IM_BILINEAR);

    FindEyes(pxRightEye, pyRightEye, pxLeftEye, pyLeftEye,
        ScaledImg, gEyeMask, x, y, EyeScale, FaceScale, 3);  // 3 is iNet i.e. eye.net
    }
}
#endif
