// $asm.cpp 1.5 milbo$ routines for active shape models
// This uses a model consisting of two stacked ASMs.
//-----------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// A copy of the GNU General Public License is available at
// http://www.r-project.org/Licenses/
//-----------------------------------------------------------------------------

#include "all.hpp"

#define PRINT_SEARCH_TIME 0

static const int  NMODELS = 2;                      // number of models
static const char *MODEL_NAME1 = "model-1.asm";     // model file names
static const char *MODEL_NAME2 = "model-2.asm";

typedef struct tSearchImages    // holds the images for one pyr level during a search
    {
    Image  Img;                 // the input image, scaled to current pyramid level
    MatVec Grads;               // image gradients
    }
tSearchImages;

bool fgExplicitPrevNext;    // specified in .asm file (TODO make these non global)
double gNormalizedProfLen;  // specified in .asm file

//-----------------------------------------------------------------------------
double GetX (double X, int iOffset, int iOrthOffset, double DeltaX, double DeltaY)
{
return X + (iOffset * DeltaX) + (iOrthOffset * DeltaY);
}

//-----------------------------------------------------------------------------
double GetY (double Y, int iOffset, int iOrthOffset, double DeltaX, double DeltaY)
{
return Y + (iOffset * DeltaY) - (iOrthOffset * DeltaX);
}

//-----------------------------------------------------------------------------
double GetPyrScale (int iLev, double PyrRatio)
{
return pow(PyrRatio, iLev);
}

//-----------------------------------------------------------------------------
bool fPointUsed (const SHAPE &Shape, int iPoint)
{
return Shape(iPoint, VX) != 0 || Shape(iPoint, VY) != 0;
}

//-----------------------------------------------------------------------------
void
SetUnusedPointsToZero (SHAPE &Shape,                // io
                       const SHAPE &RefShape)       // in
{
for (unsigned iPoint = 0; iPoint < RefShape.nrows(); iPoint++)
    if (!fPointUsed(RefShape, iPoint))
        {
        Shape(iPoint, VX) = 0;
        Shape(iPoint, VY) = 0;
        }
}

//-----------------------------------------------------------------------------
static int nGetNbrUsedPoints (const SHAPE &Shape, int nPoints)
{
int nUsed = 0;

for (int iPoint = 0; iPoint < nPoints; iPoint++)
    if (fPointUsed(Shape, iPoint))
        nUsed++;

return nUsed;
}

//-----------------------------------------------------------------------------
static int
nGetProfWidthFromModel (int iPoint, int iSub, const tAsmLev &AsmLev)
{
return nGetProfWidth(AsmLev.Profs[iPoint][iSub].ncols(), AsmLev.ProfSpecs[iPoint]);
}

//-----------------------------------------------------------------------------
// Multiply a Nx2 matrix (the InShape) by a 3x3 homogenous TransformMat
// This knows about unused landmarks, and leaves them untouched
// Returned result is the transformed InShape.

SHAPE TransformShape (const SHAPE &InShape, const Mat &TransformMat)
{
ASSERT(InShape.ncols() == 2);
ASSERT(TransformMat.ncols() == 3 && TransformMat.nrows() == 3);

SHAPE OutShape(InShape);
int iRow = InShape.nrows();
while (iRow--)
    if (fPointUsed(InShape, iRow))
        {
        MatView Row(OutShape.row(iRow));
        Row = TransformMat.mat33TimesVec2(Row);
        }

return OutShape;
}

//-----------------------------------------------------------------------------
// Align Shape to to AnchorShape using a similarity transform.
// This overwrites Shape and returns the transformation matrix i.e. the pose.
//
// This knows about unused landmarks in AnchorShape and Shape, and leaves
// them untouched. See also algorithm C.3 in Appendix C of CootesTaylor 2004
// www.isbe.man.ac.uk/~bim/Models/app_models.pdf.
//
// This is a similarity transform so the transform matrix has the form
//      a -b  Tx
//      b  a  Ty
//      0  0  1
//
// Make pWeights equal to NULL if you don't want to weight the landmarks.
// NULL is the default.

Mat AlignShape (SHAPE &Shape,                                   // io
                const SHAPE &AnchorShape, const Vec *pWeights)  // in
{
CheckSameNbrRows(Shape, AnchorShape, "AlignShape");
if (pWeights)
    ASSERT(Shape.nrows() == pWeights->nelems());
double X1 = 0, Y1 = 0, X2 = 0, Y2 = 0, W = 0, Z = 0, C1 = 0, C2 = 0;
int iRow = Shape.nrows();
while (iRow--)
    {
    double x1 = AnchorShape(iRow, VX);
    double y1 = AnchorShape(iRow, VY);
    double x2 = Shape(iRow, VX);
    double y2 = Shape(iRow, VY);

    if (x1 == 0 && y1 == 0)      // is anchor landmark unused?
        ;
    else if (x2 == 0 && y2 == 0) // is landmark unused?
        ;
    else
        {
        double w = (pWeights? (*pWeights)(iRow): 1.0);
        W  += w;
        Z  += w * (x2 * x2 + y2 * y2);
        X1 += w * x1;
        Y1 += w * y1;
        X2 += w * x2;
        Y2 += w * y2;
        C1 += w * (x1 * x2 + y1 * y2);
        C2 += w * (y1 * x2 - x1 * y2);
        }
    }
double SolnData[] = { X2, -Y2,   W,   0,
                      Y2,  X2,   0,   W,
                       Z,   0,  X2,  Y2,
                       0,   Z, -Y2,  X2 };

MatView Mat4(SolnData, 4, 4, 0);            // 4x4, tda=0

double  VecData[] = { X1, Y1, C1, C2 };
VecView Vec4(VecData, 4);

Vec Soln(SolveWithLU(Mat4, Vec4));

double TransformData[] = { Soln(0), -Soln(1), Soln(2),
                           Soln(1),  Soln(0), Soln(3),
                           0,        0,       1 };

Mat Transform(TransformData, 3, 3);

Shape = TransformShape(Shape, Transform);

return Transform;
}

//-----------------------------------------------------------------------------
// Return the Mahalanobis distance between the model profile and
// image profile at the given iPoint and offset.
// ix and iy are the offset and orthogonal offset wrt iPoint.

static double
GetProfDist (const tSearchImages &SearchImgs,         // in: all args
      const int iPoint, const int ix, const int iy,
      const tAsmLev &AsmLev, const SHAPE &Shape,
      const double SigmoidScale)
{
ASSERT(MAX_NBR_SUBPROFS == 1);  // this func supports only one subprof per prof
const unsigned iSub = 0;        // ...so this index is always 0

// It is quicker to allocate Prof statically and call dimClear() each
// time because most of the time the dimension is the same as the previous profile
// (and so we avoid a malloc and free every time this routine is called).
static Vec Prof;
Prof.dimClear(1, AsmLev.Profs[iPoint][iSub].ncols());

const int nProfWidth = nGetProfWidthFromModel(iPoint, iSub, AsmLev);
if (AsmLev.ProfSpecs[iPoint] & PROF_2d)     // two dimensional profile?
    Get2dProf(Prof,
        AsmLev.ProfSpecs[iPoint], iSub, SearchImgs.Grads,
        Shape, iPoint, ix, iy, nProfWidth, SigmoidScale);
else
    Get1dProf(Prof,
        AsmLev.ProfSpecs[iPoint], SearchImgs.Img, iPoint, ix);

// The following code is equivalent to
//      return (Prof - ModelMeanProf) * Covar * (Prof - ModelMeanProf).t();
// but is optimized for speed.

Prof -= AsmLev.Profs[iPoint][iSub]; // for efficiency, use "-=" rather than "="
const Mat *pCovar = &AsmLev.Covars[iPoint][iSub];
if (pCovar->nrows() > 1)
    return xAx(Prof, *pCovar);
else
    return Sparse_xAx(Prof, AsmLev.SparseCovars[iPoint][iSub]);
}

//-----------------------------------------------------------------------------
// limit the values of b to make sure the generated shape is plausible
// i.e. clip each b[i] to BMax * sqrt(lambda_i)

static void
LimitB (Vec &b,                                             // io
        const Vec EigVals, int nPrincipalEigs, double BMax) // in
{
int iEig;
for (iEig = 0; iEig < nPrincipalEigs; iEig++)
    {
    double Limit = BMax * sqrt(EigVals(iEig));
    if (b(iEig) < -Limit)
        b(iEig) = -Limit;
    else if (b(iEig) > Limit)
        b(iEig) = Limit;
    }
// discard unused higher components

while (iEig < int(EigVals.nrows()))
    b(iEig++) = 0.0;
}

//-----------------------------------------------------------------------------
// Return a copy of InShape conformed to the model.  In other words, generate a
// model shape that is as close as possible to InShape. InShape itself isn't changed.
//
// To match the model to Shape we need to find Pose and b in the following equation:
//
// Shape = Pose * (MeanShape + EigVecs * b)     (where = actually means approx equal)
//
// The returned shape has all the points in the model, even if InShape
// has unused points (unused points are equal to 0,0).

static SHAPE
ConformShapeToModel(Vec &b,                                         // io
        const SHAPE &InShape, const tAsmModel &Model, int iLev,     // in
        bool fShapeModelFinalIter)                                  // in
{
Vec MeanShape(Model.AsmLevs[iLev].MeanShape);
SetUnusedPointsToZero(MeanShape, InShape);

// For calculations below we need to see shapes (nrows x 2) as vectors (1 x 2*nrows).
// Note that this is a "view" so if you change MeanShapeAsVec you
// are changing MeanShape too, and vice versa.

VecView MeanShapeAsVec(MeanShape.viewAsCol());

// find y, the model shape that best fits InShape

SHAPE Shape(InShape);
int nEigs = Model.AsmLevs[iLev].nEigs;
double BMax = Model.AsmLevs[iLev].BMax;
if (fShapeModelFinalIter)   // if final iter in main ASM search loop then loosen up the model
    {
    ASSERT(Model.AsmLevs[iLev].nEigsFinal > 0);
    ASSERT(Model.AsmLevs[iLev].BMaxFinal > 0);
    nEigs = Model.AsmLevs[iLev].nEigsFinal;
    BMax = Model.AsmLevs[iLev].BMaxFinal;
    }
SHAPE x(InShape.nrows(), 2);
x.viewAsCol() = MeanShapeAsVec + Model.EigVecs * b; // generate a model shape x
Mat Pose(AlignShape(x, InShape, NULL));
SHAPE y(TransformShape(InShape, Pose.inverse()));   // project InShape into model space

// update model params b to match y

b = Model.EigInverse * (y.viewAsCol() - MeanShapeAsVec);
LimitB(b, Model.AsmLevs[iLev].EigVals, nEigs, BMax); // ensure we stay within model limits

// generate Shape from the model using our calculated b, and align Shape to InShape

Shape.viewAsCol() = Model.EigVecs * b;
Shape = TransformShape(Model.AsmLevs[iLev].MeanShape + Shape, Pose);

return Shape;
}

//-----------------------------------------------------------------------------
// Return offsets (from iPoint) ixBest and iyBest at the best profile match.
// Table to explain what x and y directions mean:
//
//             1D                       2D
//    x    offset along whisker     horiz offset
//    y    always 0                 vert offset

static void
FindBestMatchingProf (int &ixBest, int &iyBest,                         // out
    int iPoint, const SHAPE& Shape, const tLand LandTab[],              // in
    int nMaxOffset, tSearchImages &SearchImgs, const tAsmLev &Model,    // in
    double SigmoidScale)                                                // in
{
int nyMaxOffset = 0;
unsigned ProfSpec = Model.ProfSpecs[iPoint];
if (ProfSpec & PROF_2d)                         // two dimensional profile?
    nyMaxOffset = nMaxOffset;                   // if so, we need to offset in y dir also
else
    PrepareProf1D(SearchImgs.Img, Shape, ProfSpec, LandTab, iPoint,
        nGetProfWidthFromModel(iPoint, 0, Model) + 2 * nMaxOffset,
        0, 0, fgExplicitPrevNext);

double BestFit = DBL_MAX;
ixBest = 0, iyBest = 0;

for (int iy = -nyMaxOffset; iy <= nyMaxOffset; iy++)
    for (int ix = -nMaxOffset; ix <= nMaxOffset; ix++)
        {
        double Fit = GetProfDist(SearchImgs, iPoint, ix, iy,
                        Model, Shape, SigmoidScale);

        // Test for a new best fit. We test using "<=" instead of just "<"
        // so if there is an exact match then ixBest=0 i.e. no change.

        if ((ix <= 0 && iy <= 0)? Fit <= BestFit:  Fit < BestFit)
            {
            ixBest = ix;
            iyBest = iy;
            BestFit = Fit;
            }
        }
}

//-----------------------------------------------------------------------------
// Update Shape to SuggestedShape by matching the image profile to
// the model profile at each landmark.
// Returns number of landmarks that are "good".

static int
GetSuggestedShape (SHAPE& SuggestedShape,                           // io
    const SHAPE& Shape, int nMaxOffset, tSearchImages &SearchImgs,  // in
    const tAsmLev &Model, const tLand LandTab[],                    // in
    double SigmoidScale)                                            // in
{
int nGoodLandmarks = 0;
int ixBest, iyBest;
int nPoints = SuggestedShape.nrows();

for (int iPoint = 0; iPoint < nPoints; iPoint++)
    if (fPointUsed(Shape, iPoint))
        {
        FindBestMatchingProf(ixBest, iyBest,
            iPoint, Shape, LandTab, nMaxOffset, SearchImgs, Model, SigmoidScale);

        // set SuggestedShape(iPoint) to best offset from current position

        if (Model.ProfSpecs[iPoint] & PROF_2d)                  // two dimensional profile?
            {
            // x,y orthogonal to image sides (not to whisker)

            SuggestedShape(iPoint, VX) = Shape(iPoint, VX) - ixBest;
            SuggestedShape(iPoint, VY) = Shape(iPoint, VY) - iyBest;
            if (ABS(ixBest) + ABS(iyBest) <= nMaxOffset)
                nGoodLandmarks++;
            }
        else    // one dimensional profile: must move point along the whisker
            {
            double DeltaX = 0, DeltaY = 0;
            if (ixBest || iyBest)
                GetProfStepSize(DeltaX, DeltaY,
                    Shape, iPoint, 0, LandTab, fgExplicitPrevNext);

            SuggestedShape(iPoint, VX) = GetX(Shape(iPoint, VX),
                                            ixBest, iyBest, DeltaX, DeltaY);
            SuggestedShape(iPoint, VY) = GetY(Shape(iPoint, VY),
                                            ixBest, iyBest, DeltaX, DeltaY);
            if (ABS(ixBest) <= nMaxOffset/2)
                nGoodLandmarks++;
            }
        }
return nGoodLandmarks;
}

//-----------------------------------------------------------------------------
// do an ASM search at one level in the image pyramid

static void
AsmLevSearch (SHAPE &Shape,     // io
    tSearchImages &SearchImgs,  // in
    const tAsmModel &Model,     // in
    int iLev,                   // in
    const tLand LandTab[])      // in
{
int   iter = 0, nGoodLandmarks = 0;
SHAPE SuggestedShape(Shape);    // shape after profile matching

// The shape params, initialized to 0.  The original formulation called for
// this to be set to 0 each time we run the model but we get slightly
// better results if we remember the shape params from the previous run.
// Thus this is outside the loop.

Vec b(Model.EigVecs.nrows());

int nUsedPoints = nGetNbrUsedPoints(Shape, Shape.nrows());

while ((iter < Model.nMaxSearchIters) &&
        (nGoodLandmarks <= (Model.nQualifyingDisplacments * nUsedPoints)/100))
    {
    // estimate the best SuggestedShape by profile matching the landmarks in Shape

    nGoodLandmarks = GetSuggestedShape(SuggestedShape, Shape, Model.nMaxOffset, SearchImgs,
                            Model.AsmLevs[iLev], LandTab, Model.SigmoidScale);

    // align SuggestedShape to the shape model, put result in Shape

    bool fFinalIter = (iter == Model.nMaxSearchIters - 1);
    Shape = ConformShapeToModel(b, SuggestedShape, Model, iLev, fFinalIter);

    iter++;
    }
}

//-----------------------------------------------------------------------------
// Accumulate the results of the current search (in Shape) into the
// combined results for all searches (in CombinedShape)
// This basically copies Shape to CombinedShape but has to deal with
// the fact that these shapes may be different sizes.

static void
CombineSearchResults (
    SHAPE &CombinedShape,           // io: best shape formed by combining all searches
    int iModel, const SHAPE &Shape) // in: results of current search
{
if (iModel == 0)
    CombinedShape = Shape;
else
    {
    unsigned nPoints = CombinedShape.nrows();
    ASSERT(Shape.nrows() >= nPoints);
    for (unsigned iPoint = 0; iPoint < nPoints; iPoint++)
        {
        CombinedShape(iPoint, VX) = Shape(iPoint, VX);
        CombinedShape(iPoint, VY) = Shape(iPoint, VY);
        }
    }
}

//-----------------------------------------------------------------------------
// init Models from .asm files and global defines

static void
InitAsmModels (tAsmModel Models[],  // out: two ASM models
            const char *sDataDir)   // in: dir holding .asm files
{
static bool fInitialized = false;
if (!fInitialized)                  // Models not yet initialized?
    {
    ASSERT(NMODELS == 2);           // this function only supports 2 models

    char sFile[SLEN];
    sprintf(sFile, "%s/%s", sDataDir, MODEL_NAME1);
    InitAsm(Models[0], sFile);

    sprintf(sFile, "%s/%s", sDataDir, MODEL_NAME2);
    InitAsm(Models[1], sFile);

    Models[1].nStartLev = 1;        // second model starts at pyramid level 1

    fInitialized = true;
    }
}

//-----------------------------------------------------------------------------
SHAPE                               // results returned as a SHAPE
AsmSearch (const RgbImage &RgbImg,  // in: find face features in this image
    const bool fViolaJones,         // in: true to use VJ detector, else Rowley
    const char sImageFile[],        // in: file path for RgbImg, for err msgs
    const char sDataDir[])          // in: directory for data files
{
Image Img;
ConvertRgbImageToGray(Img, RgbImg);

static tAsmModel Models[NMODELS];
InitAsmModels(Models, sDataDir);

#if PRINT_SEARCH_TIME
clock_t StartTime;
static double TotalDetTime, TotalAsmTime;
static clock_t nImages;
nImages++;
StartTime = clock();
#endif

SHAPE Shape;
GetStartShapeFromGlobalDet(Shape,
    fViolaJones, Img, sImageFile, Models[0].FileMeanShape, sDataDir);

#if PRINT_SEARCH_TIME
TotalDetTime += double(clock() - StartTime) / CLOCKS_PER_SEC;
lprintf("MeanFaceDetTime %.3f ", TotalDetTime / nImages);
StartTime = clock();
#endif
double OrgFaceWidth = Shape(MRJaw2, VX) - Shape(MLJaw2, VX);

// CombinedShape is created by combining shapes from each model search.  It
// has the same nbr of points as Models[0].FileMeanShape, so if Models[1]
// has extra points they are discarded.

SHAPE CombinedShape;

for (int iModel = 0; iModel < NMODELS; iModel++)
    {
    tAsmModel *pModel = &Models[iModel];
    fgExplicitPrevNext = pModel->fExplicitPrevNext; // init nasty globals
    gNormalizedProfLen = pModel->NormalizedProfLen;
    if (iModel != 0)
        GetStartShapeFromPreviousSearch(Shape, CombinedShape, pModel->FileMeanShape);

    // Scale Shape and Img, so face width is a constant nScaledFaceWidth.
    // We use StartShape to approximately determine face width.

    double ImageScale = double(pModel->nScaledFaceWidth) / OrgFaceWidth;
    SHAPE Shape1(Shape * ImageScale);   // working shape
    Image Img1(Img);                    // working Img
    int nNewWidth  = int(Img1.width * ImageScale);
    int nNewHeight = int(Img1.height * ImageScale);
    ScaleImage(Img1, nNewWidth, nNewHeight, IM_BILINEAR);

    // dimKeep is needed when this model has different nbr of landmarks from previous model
    Shape1.dimKeep(pModel->nPoints, 2);
    SetUnusedPointsToZero(Shape1, pModel->AsmLevs[0].MeanShape);
    int nStartLev = pModel->nStartLev;
    Shape1 /= GetPyrScale(nStartLev, pModel->PyrRatio);
    for (int iLev = nStartLev; iLev >= 0; iLev--)   // for each level in the image pyramid
        {
        double PyrScale = GetPyrScale(iLev, pModel->PyrRatio);
        tSearchImages SearchImgs;       // the images used during search
        SearchImgs.Img = Img1;          // SearchImgs.Img gets scaled to this pyramid level
        ReduceImage(SearchImgs.Img, PyrScale, pModel->PyrReduceMethod);
        InitGradsIfNeeded(SearchImgs.Grads,         // get SearchImgs.Grads
            pModel->AsmLevs[iLev].ProfSpecs, SearchImgs.Img, Shape1.nrows());

        AsmLevSearch(Shape1, SearchImgs, Models[iModel], iLev, gLandTab);

        if (iLev != 0)      // use best shape from this iter as starting point for next
            Shape1 *= pModel->PyrRatio;
        }
    CombineSearchResults(CombinedShape, iModel, Shape1);
    CombinedShape = CombinedShape / ImageScale;     // descale back to original size
    }
#if PRINT_SEARCH_TIME
TotalAsmTime += double(clock() - StartTime) / CLOCKS_PER_SEC;
lprintf("MeanAsmTime %.3f ", TotalAsmTime / nImages);
#endif
return CombinedShape;
}
