// $asmfile.cpp 1.5 milbo$ routines to read .asm files
//-----------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// A copy of the GNU General Public License is available at
// http://www.r-project.org/Licenses/
//-----------------------------------------------------------------------------

#include "all.hpp"

//-----------------------------------------------------------------------------
static void ReadInt (int &i, const char si[], int iMin, int iMax, FILE *pAsmFile)
{
char s[SLEN], sToken[SLEN];
if (!Fgets(s, SLEN-1, pAsmFile))
    Err("Can't read %s from ASM file", si);
i = INT_MAX;
if (2 != sscanf(s, "%s %d", sToken, &i) || 0 != strcmp(sToken, si))
    Err("Can't read %s from ASM file", si);
if (i < iMin || i > iMax)
    Err("Bad value %d for %s in ASM file header", i, si);
}

//-----------------------------------------------------------------------------
static void ReadDouble (double &d, const char sd[], double dMin, double dMax, FILE *pAsmFile)
{
char s[SLEN], sToken[SLEN];
if (!Fgets(s, SLEN-1, pAsmFile))
    Err("Can't read %s from ASM file", sd);
float Temp = FLT_MAX;    // float needed for %g (for MS compiler at least)
if (2 != sscanf(s, "%s %g", sToken, &Temp) || 0 != strcmp(sToken, sd))
    Err("Can't read %s from ASM file", sd);
if (Temp < dMin || Temp > dMax)
    Err("Bad value %g for %s in ASM file header", Temp, sd);
d = Temp;
}

//-----------------------------------------------------------------------------
static void ReadBool (bool &b, const char sb[], FILE *pAsmFile)
{
char s[SLEN], sToken[SLEN];
if (!Fgets(s, SLEN-1, pAsmFile))
    Err("Can't read %s from ASM file", sb);
int Temp = -1;      // needed if sizeof(bool) != sizeof(int) which is compiler dependent
if (2 != sscanf(s, "%s %d", sToken, &Temp) || 0 != strcmp(sToken, sb))
    Err("Can't read %s from ASM file", sb);
if (Temp != 0 && Temp != 1)
    Err("Bad value %d for %s in ASM file header", Temp, sb);
b = (Temp == 1);
}

//-----------------------------------------------------------------------------
// check first line of .asm file

static void
CheckHeader (const char sAsmFile[], FILE *pAsmFile) // in
{
// first line of asm file "ASM0 [path/filename.asm]"

char s[SLEN], sFile[SLEN];
Fgets(s, SLEN-1, pAsmFile);

if (strncmp(s, "ASM", 3) != 0)          // check ASM file magic number
    Err("%s is not an ASM file (first three chars are not \"ASM\")", sAsmFile);

int nVersion;
if (sscanf(s+3, "%d %s", &nVersion, sFile) != 2)
    Err("Bad format ASM file header: %s", s);

if (nVersion != CONF_nAsmFileVersion)
    Err("Bad version number %d in ASM file header", nVersion);
}

//-----------------------------------------------------------------------------
// Read and check header of a level within an .asm file
// No results are kept, this just checks that the header is good

static void
CheckLevelHeader (const char sAsmFile[], FILE *pAsmFile, int iLev) // in: all
{
// expect line to look like this: Lev 0 FirstShape xExtent 182 yExtent 215

char s[SLEN], sLev[SLEN], sFirstShape[SLEN];
int nLev;

if (!Fgets(s, SLEN-1, pAsmFile))
    Err("Can't read level %d header from %s", iLev, sAsmFile);

if (sscanf(s, "%s %d %s", sLev, &nLev, sFirstShape) != 3)
    Err("Can't read level %d header from %s, line is %s", iLev, sAsmFile, s);

if (strcmp(sLev, "Lev") != 0 || strcmp(sFirstShape, "FirstShape") != 0)
    Err("Can't read level %d header from %s, line is %s", iLev, sAsmFile, s);

if (nLev != iLev)
    Err("Illegal level %d in level header from %s, expected %d, line is %s",
        nLev, sAsmFile, iLev, s);
}

//-----------------------------------------------------------------------------
// This duplicates the functionality of Mat.read but for cvec<unsigned>.
// TODO replace this function and ReadIntVec, duplicated code?

static void
ReadProfSpecs (cvec<unsigned> &ProfSpecs,                   // out
      int nPoints, const char sAsmFile[], FILE *pAsmFile)   // in
{
char s[SLEN];

// header "{ 68 1"

if (!Fgets(s, SLEN-1, pAsmFile))
    Err("Can't read ProfSpecs header from %s", sAsmFile);
char sBrace[SLEN];
int nrows = -1, ncols = -1;
if (sscanf(s, "%s %d %d", sBrace, &nrows, &ncols) != 3 ||
        sBrace[0] != '{' || sBrace[1] != 0)
    Err("Bad format ProfSpecs header: %s", s);

if (nrows != nPoints || ncols != 1)
    Err("Bad number of rows or cols in ProfSpecs header, got %d %d expected %d 1",
        nrows, ncols, nPoints);

// body, lines like "1000000"

ProfSpecs.resize(nPoints);
for (int iPoint = 0; iPoint < nPoints; iPoint++)
    {
    if (!Fgets(s, SLEN-1, pAsmFile))
        Err("Can't read ProfSpecs from %s", sAsmFile);
    unsigned ProfSpec = unsigned(-1);
    if (sscanf(s, "%x", &ProfSpec) != 1 || ProfSpec == unsigned(-1))
        Err("Can't read ProfSpecs from %s", sAsmFile);
    if (nSubProfsForProfSpec(ProfSpec) > MAX_NBR_SUBPROFS)
        Err("Profile %x in %s has more than %d sub profiles",
            ProfSpec, sAsmFile, MAX_NBR_SUBPROFS);
    ProfSpecs[iPoint] = ProfSpec;
    }

// footer "}"

if (!Fgets(s, SLEN-1, pAsmFile))
    Err("Can't read ProfSpecs footer from %s", sAsmFile);
if (s[0] != '}')
    Err("Bad format ProfSpecs footer, expected closing brace, got %s", s);
}

//-----------------------------------------------------------------------------
// This gets called AFTER initializing at least following for Asm[0]:
//      nPoints EigVals EigVecs EigInverse nEigs

static void
ReadAsmLev (tAsmLev &AsmLev,                // out
    int iLev,                               // in
    int nPoints,                            // in
    const char sAsmFile[], FILE *pAsmFile)  // in
{
CheckLevelHeader(sAsmFile, pAsmFile, iLev);

ReadProfSpecs(AsmLev.ProfSpecs, nPoints, sAsmFile, pAsmFile);

AsmLev.Covars.resize(nPoints);
AsmLev.SparseCovars.resize(nPoints);

int iPoint;
for (iPoint = 0; iPoint < nPoints; iPoint++)
    ReadMatVec(&AsmLev.Covars[iPoint], sAsmFile, pAsmFile);

AsmLev.Profs.resize(nPoints);

for (iPoint = 0; iPoint < nPoints; iPoint++)
    ReadMatVec(&AsmLev.Profs[iPoint], sAsmFile, pAsmFile);

// convert sparse covar matrices from Mat to SparseMat, where necessary

ASSERT(MAX_NBR_SUBPROFS == 1);  // this func supports only one subprof per prof
const unsigned iSub = 0;        // ...so this index is always 0
for (iPoint = 0; iPoint < nPoints; iPoint++)
    if (fSparseMat(AsmLev.Covars[iPoint][iSub]))
        {
        if (AsmLev.SparseCovars[iPoint].size() < 1)
            AsmLev.SparseCovars[iPoint].resize(1);

        CopyMatToSparseMat(AsmLev.SparseCovars[iPoint][iSub],AsmLev.Covars[iPoint][iSub]);

        // TODO We set the dimensions here to 1,1 not 0,0 so calls to nrows()
        // later don't fail -- should be revisited for a cleaner interface?

        AsmLev.Covars[iPoint][iSub].dim(1, 1);  // mark Covars as unneeded
        }
}

//-----------------------------------------------------------------------------
// return true if there are any 2D profs in the given model level

static bool
fTwoDimProfs (const tAsmLev &Model, int nPoints)    // in: all
{
for (int iPoint = 0; iPoint < nPoints; iPoint++)
    if (Model.ProfSpecs[iPoint] & PROF_2d)          // two dimensional profile?
        return true;

return false;
}

//-----------------------------------------------------------------------------
static void
Init_nEigs_bMax(tAsmModel &Asm,                 // io
                int iLev, int fTwoDimProfsLev0) // in
{
if (iLev == 0)
    {
    if (fTwoDimProfsLev0)
        {
        if (CONF_nEigsLev0_2d)
            {
            Asm.AsmLevs[0].nEigs = CONF_nEigsLev0_2d;
            Asm.AsmLevs[0].BMax = CONF_BMaxLev0_2d;
            }
        else
            {
            Asm.AsmLevs[0].nEigs = CONF_nEigs;
            Asm.AsmLevs[0].BMax = CONF_BMax;
            }
        }
    else
        {
        if (CONF_nEigsLev0)
            {
            Asm.AsmLevs[0].nEigs = CONF_nEigsLev0;
            Asm.AsmLevs[0].BMax = CONF_BMaxLev0;
            }
        else
            {
            Asm.AsmLevs[0].nEigs = CONF_nEigs;
            Asm.AsmLevs[0].BMax = CONF_BMax;
            }
        }
    }
else    // not iLev0
    {
    Asm.AsmLevs[iLev].nEigs = CONF_nEigs;
    Asm.AsmLevs[iLev].BMax = CONF_BMax;
    }
if (CONF_nEigsFinal)
    {
    Asm.AsmLevs[iLev].nEigsFinal = CONF_nEigsFinal;
    Asm.AsmLevs[iLev].BMaxFinal = CONF_BMaxFinal;
    }
else
    {
    Asm.AsmLevs[iLev].nEigsFinal = CONF_nEigs;
    Asm.AsmLevs[iLev].BMaxFinal = CONF_BMax;
    }
}

//-----------------------------------------------------------------------------
// init Asm by reading a .asm file and by looking at global CONF_ settings

void
InitAsm (tAsmModel &Asm,        // out
         const char sAsmFile[]) // in
{
FILE *pAsmFile = Fopen(sAsmFile, "r");

// header

CheckHeader(sAsmFile, pAsmFile);

// prolog
// The magic numbers like 1000 below are arbitrary and intended to catch gross errs

int nLevs;
bool fBilinearRescale;

ReadInt(Asm.nPoints,              "Points",              1, MAX_NBR_LANDMARKS, pAsmFile);
ReadInt(nLevs,                    "Levs",                1, MAX_NBR_LEVS, pAsmFile);
ReadDouble(Asm.PyrRatio,          "PyramidRatio",        1, 1000, pAsmFile);
ReadInt(Asm.PyrReduceMethod,      "PyramidReduceMethod", IM_NEAREST_PIXEL, IM_AVERAGE_ALL,
                                                         pAsmFile);
ReadBool(Asm.fExplicitPrevNext,   "ExplicitPrevNext",    pAsmFile);
ReadDouble(Asm.NormalizedProfLen, "NormalizedProfLen",   1, 1000, pAsmFile);
ReadInt(Asm.nScaledFaceWidth,     "ScaledFaceWidth",     0, 1000, pAsmFile);
ReadBool(fBilinearRescale,        "BilinearRescale",     pAsmFile);
ReadInt(Asm.nTrim,                "TrimCovar",           0, 1000, pAsmFile);
ReadDouble(Asm.SigmoidScale,      "SigmoidScale",        0.001, 1e10, pAsmFile);

Asm.nStartLev = nLevs - 1;

// eigen values and vectors

Asm.AsmLevs[0].EigVals.read(sAsmFile, pAsmFile);
Asm.EigVecs.read(sAsmFile, pAsmFile);
Asm.EigInverse = Asm.EigVecs.t(); // inverse of EigVecs (EigVecs is orthog, so transp=inverse)

// file mean shape

Asm.AsmLevs[0].MeanShape.read(sAsmFile, pAsmFile);

ASSERT(unsigned(Asm.nPoints) == Asm.AsmLevs[0].MeanShape.nrows());
ASSERT(2 * unsigned(Asm.nPoints) == Asm.EigVecs.nrows());

// level specific data

int iLev;
for (iLev = nLevs-1; iLev >= 0; iLev--)     // for each level in the image pyramid
    ReadAsmLev(Asm.AsmLevs[iLev], iLev, Asm.nPoints, sAsmFile, pAsmFile);

fclose(pAsmFile);

Asm.FileMeanShape = Asm.AsmLevs[0].MeanShape;

// initializations from global CONF_ #defines

bool fTwoDimProfsLev0 = fTwoDimProfs(Asm.AsmLevs[0], Asm.nPoints);
if (fTwoDimProfsLev0)
    {
    Asm.nMaxOffset = CONF_nPixSearch2d;
    Asm.nMaxSearchIters = CONF_nMaxSearchIters2d;
    Asm.nQualifyingDisplacments = CONF_nQualifyingDisplacments2d;
    }
else
    {
    Asm.nMaxOffset = CONF_nPixSearch;
    Asm.nMaxSearchIters = CONF_nMaxSearchIters;
    Asm.nQualifyingDisplacments = CONF_nQualifyingDisplacments;
    }
for (iLev = 0; iLev < nLevs; iLev++)    // for each level in the image pyramid
    {
    Init_nEigs_bMax(Asm, iLev, fTwoDimProfsLev0);

    Asm.AsmLevs[iLev].EigVals = Asm.AsmLevs[0].EigVals /
                                    pow(Asm.PyrRatio * Asm.PyrRatio, iLev);

    // In theory we shouldn't have to scale MeanShape here because
    // ConformShapeToModel takes care of scaling. In practice we get better
    // results when we do scale here.  TODO Look into this.

    Asm.AsmLevs[iLev].MeanShape = Asm.AsmLevs[0].MeanShape /
                                    GetPyrScale(iLev, Asm.PyrRatio);
    }
}
