// $imshape.cpp 1.5 milbo$ routines for using shapes in images
//-----------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// A copy of the GNU General Public License is available at
// http://www.r-project.org/Licenses/
//-----------------------------------------------------------------------------

#include "all.hpp"

// if true, skip lines connecting extra eye points >= 68 in DrawShape
// i.e. if true, draw the shape using the same points as an XM2VTS shape

static const bool SKIP_EXTRA_EYE_LINES_IN_DISPLAYED_SHAPE = true;

//-----------------------------------------------------------------------------
static void
DrawPoint (RgbImage &Img,                           // io
    int ix, int iy,                                 // in
    unsigned Red, unsigned Green, unsigned Blue)    // in
{
int width = Img.width, height = Img.height;
ix += width / 2;       // convert shape to image coords
iy += height / 2;
if (ix >= 0 && ix < width && iy >= 0 && iy < height)
    {
    Img(ix, iy).Red = Red;
    Img(ix, iy).Green = Green;
    Img(ix, iy).Blue = Blue;
    }
}

//-----------------------------------------------------------------------------
// This function uses a Bresenham-like algorithm to draw a line from x0,y0 to x1,y1.
// Based on "Bresenham-based supercover line algorithm" by
// Eugen Dedu www.ese-metz.fr/~dedu/projects/bresenham.

static void
DrawLine (RgbImage &Img,                            // io
    int x0, int y0, int x1, int y1,                 // in
    unsigned Red, unsigned Green, unsigned Blue)    // in
{
int i;               // loop counter
int yStep, xStep;    // the step on y and x axis
int Error;           // the error accumulated during the increment
int ddy, ddx;
int x = x0, y = y0;
int dx = x1 - x0;
int dy = y1 - y0;

DrawPoint(Img, x, y, Red, Green, Blue);     // first point

if (dy < 0)
    {
    yStep = -1;
    dy = -dy;
    }
else
    yStep = 1;

if (dx < 0)
    {
    xStep = -1;
    dx = -dx;
    }
else
    xStep = 1;

ddy = 2 * dy;
ddx = 2 * dx;

if (ddx >= ddy)                 // first octant (0 <= slope <= 1)
    {
    Error = dx;                 // start in the middle of the square
    for (i = 0; i < dx; i++)    // do not use the first point (already done)
        {
        x += xStep;
        Error += ddy;
        if (Error > ddx)        // increment y if AFTER the middle (>)
            {
            y += yStep;
            Error -= ddx;
            }
        DrawPoint(Img, x, y, Red, Green, Blue);
        }
    }
else
    {
    Error = dy;
    for (i = 0; i < dy; i++)
        {
        y += yStep;
        Error += ddx;
        if (Error > ddy)
            {
            x += xStep;
            Error -= ddy;
            }
        DrawPoint(Img, x, y, Red, Green, Blue);
        }
    }
}

//-----------------------------------------------------------------------------
// draw Shape in the image Img in the specified Color

void
DrawShape (RgbImage &Img,                        // io
           const SHAPE &Shape, unsigned Color)   // in
{
unsigned Red   = (Color >> 16) & 0xff;
unsigned Green = (Color >> 8) & 0xff;
unsigned Blue  = Color & 0xff;

int iNext = 0;  // next point
int nPoints = Shape.nrows();
for (int iPoint = 0; iPoint < nPoints; iPoint++)
    {
    if (!fPointUsed(Shape, iPoint))
        continue;

    if (iPoint < nPoints-1)
        iNext = iPoint + 1;
    while (iNext < nPoints && !fPointUsed(Shape, iNext))
        iNext++;

    int ix = iRound(Shape(iPoint, VX));
    int iy = iRound(Shape(iPoint, VY));
    if (iNext < nPoints &&
            (SKIP_EXTRA_EYE_LINES_IN_DISPLAYED_SHAPE ||
            iNext < MLEye0 || iNext > MREye7))
        {
        DrawLine(Img, ix, iy,
            iRound(Shape(iNext, VX)), iRound(Shape(iNext, VY)),
            Red, Green, Blue);
        }
    DrawPoint(Img, ix, iy, Red, Green, Blue);
    }
}

//-----------------------------------------------------------------------------
// crop the image so the shape fills the whole image

void
CropImageToShape (RgbImage &Img,        // io
                 const SHAPE &Shape)    // in
{
// increase the following margins if you want more room around the shape
double xMargin = 20, yMargin = 30;

int nLeftCrop   = int(MAX(0, Img.width/2  + Shape.col(VX).minElem() - xMargin));
int nRightCrop  = int(MAX(0, Img.width/2  - Shape.col(VX).maxElem() - xMargin));
int nTopCrop    = int(MAX(0, Img.height/2 - Shape.col(VY).maxElem() - yMargin));
int nBottomCrop = int(MAX(0, Img.height/2 + Shape.col(VY).minElem() - yMargin));

CropRgbImage(Img, nTopCrop, nBottomCrop, nLeftCrop, nRightCrop, IM_WIDTH_DIVISIBLE_BY_4);
}
