// $prof.cpp 1.5 milbo$ routines for ASM profiles
//-----------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// A copy of the GNU General Public License is available at
// http://www.r-project.org/Licenses/
//-----------------------------------------------------------------------------

#include "all.hpp"

//-----------------------------------------------------------------------------
int
nGetProfWidth (int ncols, unsigned ProfSpec)    // in: all
{
if ((ProfSpec & PROF_2d) == 0)      // one dimensional profile?
    return ncols;
DASSERT(SQR(sqrt(ncols)) == ncols); // check that ncols is indeed a squared integer
return int(sqrt(ncols));
}

//-----------------------------------------------------------------------------
// Each profile can have up to 3 sub profs (see prof.hpp).  This says how many.

int
nSubProfsForProfSpec (unsigned ProfSpec)        // in: all
{
#if MAX_NBR_SUBPROFS == 1
    return 1;
#else
    ASSERT(CONF_nMaxSubProfs <= 3);     // this routine can count up to 3
    ASSERT((ProfSpec & PROF_2d) || (ProfSpec & 0x00ffff00) == 0);
                                        // 1d profiles can have only one subprofile
    unsigned nSubProfs = 1;
    if (ProfSpec & 0xff00)
        {
        nSubProfs++;
        if (ProfSpec & 0xff0000)
            nSubProfs++;
        }
    return nSubProfs;
#endif
}

//-----------------------------------------------------------------------------
static unsigned
GetSubProfSpec (unsigned Prof, unsigned iSub)   // in: all
{
DASSERT(iSub < MAX_NBR_SUBPROFS);
DASSERT(iSub < CONF_nMaxSubProfs);
return (((Prof >> (iSub * 8)) & 0xff) | (Prof & 0xff000000));
}

//-----------------------------------------------------------------------------
// Return normalized vector bisector of three ordered points.
// We rotate the vector by OffsetAngle (in radians).
// Make OffsetAngle=0 to get standard bisector.

static Vec
GetBisector (const Vec &Prev, const Vec &This,          // in: all
             const Vec &Next, double OffsetAngle)
{
static Vec p(2, ROWVEC), q(2, ROWVEC);  // static to reduce number of mallocs

ASSERT(!(Prev(VX) == 0 && Prev(VY) == 0));
ASSERT(!(This(VX) == 0 && This(VY) == 0));
ASSERT(!(Next(VX) == 0 && Next(VY) == 0));

static Vec d0; d0.assign(This); d0 -= Prev; d0.normalizeMe();

// rotate 90 degrees + OffsetAngle

p(VX) = sin(OffsetAngle)  * d0(VX) + cos(OffsetAngle) * d0(VY);
p(VY) = -cos(OffsetAngle) * d0(VX) + sin(OffsetAngle) * d0(VY);

static Vec d1; d1.assign(Next); d1 -=This; d1.normalizeMe();
q(VX) = sin(OffsetAngle)  * d1(VX) + cos(OffsetAngle) * d1(VY);
q(VY) = -cos(OffsetAngle) * d1(VX) + sin(OffsetAngle) * d1(VY);

Vec d(p); d += q; d.normalizeMe();

if (ABS(d(VX)) < 1e-10 && ABS(d(VY)) < 1e-10)   // are points Prev and Next in same line?
    return (This - Prev).normalize();           // yes, avoid numerical issues

return d;
}

//-----------------------------------------------------------------------------
// Get pixel in Img at ix and iy (which are SHAPE coords), with index range forcing.
//
// SHAPE coords: 0,0 is center of image (makes it easier to deal with various size images)
//               image spans -width/2 to width/2, -height/2 to height/2

static int
iGetPixel (const Image &Img, int ix, int iy)    // in: all
{
const int width = Img.width; 
const int height = Img.height;
ix += width/2;     // convert SHAPE to Image coords
iy += height/2;
if (ix < 0)
    ix = 0;
else if (ix > width-1)
    ix = width-1;
if (iy < 0)
    iy = 0;
else if (iy > height-1)
    iy = height-1;

return Img(ix, (height - 1 - iy)); // height-1-iy min is 0 when iy at max=height-1
}

//-----------------------------------------------------------------------------
// Get normalized step size so we take single pixel steps.
// So either DeltaX or DeltaY will be +-1, and the other will be smaller than +-1.
//
// ProfAngle is the angle of the whisker relative to the orthogonal.
// Use 0 to get a conventional orthogonal whisker.
//
// *pDeltaX is along whisker, *pDeltaY is orthogonal to whisker

void
GetProfStepSize (double &DeltaX, double &DeltaY,                        // out
                 const SHAPE &Shape, int iPoint, double ProfAngle,      // in
                 const tLand LandTab[], const bool fExplicitPrevNext)   // in
{
ASSERT(ProfAngle == 0); // not supported in this version of GetProfSize
ASSERT(fPointUsed(Shape, iPoint));
const int nPoints = Shape.nrows();

// Use LandTab prev and next points if LandTab explictly specifies them.
//
// If LandTab doesn't explicitly specify prev and next landmarks, use the
// one-less and one-more index in Shape to select prev and next landmarks
// i.e. the classic Cootes method.

int iPrev, iNext;
static Vec PrevRow; PrevRow.dim(Shape.ncols(), ROWVEC);
static Vec NextRow; NextRow.dim(Shape.ncols(), ROWVEC);

if (fExplicitPrevNext)
    {
    iPrev = LandTab[iPoint].iPrev;
    if (iPrev < 0)
        iPrev = (iPoint + nPoints - 1) % nPoints;
    PrevRow = Shape.row(iPrev);
    ASSERT(fPointUsed(Shape, iPrev));
    iNext = LandTab[iPoint].iNext;
    if (iNext < 0)
        iNext = (iPoint + 1) % nPoints;
    NextRow = Shape.row(iNext);
    ASSERT(fPointUsed(Shape, iNext));
    }
else // classic Cootes method, but skip over unused points in Shape
    {
    iPrev = iPoint;
    do
        if (--iPrev < 0)
            iPrev = nPoints-1;
    while (!fPointUsed(Shape, iPrev));
    PrevRow = Shape.row(iPrev);

    iNext = iPoint;
    do
        if (++iNext >= nPoints)
            iNext = 0;
    while (!fPointUsed(Shape, iNext));
    NextRow = Shape.row(iNext);
    }
static Vec Whisker;
Whisker.assign(GetBisector(PrevRow, Shape.row(iPoint), NextRow, ProfAngle));

DeltaX = Whisker(VX);
DeltaY = Whisker(VY);
const double AbsDeltaX = ABS(DeltaX);
const double AbsDeltaY = ABS(DeltaY);
if (AbsDeltaX >= AbsDeltaY)
    {
    DeltaY /= AbsDeltaX;
    DeltaX /= AbsDeltaX;
    }
else
    {
    DeltaX /= AbsDeltaY;
    DeltaY /= AbsDeltaY;
    }
}

//-----------------------------------------------------------------------------
// variables for PrepareProf and GetProf routines

// max number of elems in a profile including nMaxOffset on each end
static const int MAX_PROF_WIDTH_1D = 50;

static Vec          gProf(MAX_PROF_WIDTH_1D, MAX_PROF_WIDTH_1D);
static int          ngProfWidth;
static int          igProfPoint;    // for sanity checking
static const byte   *pgProfImage;   // ditto
static unsigned     gSubProfSpec;   // ditto

//-----------------------------------------------------------------------------
// If you want multiple 1D profiles along one whisker, then
//      i.  call PrepareProf1D once with nProfWidth set to the length of the entire prof
//      ii. call Get1dProf multiple times, incrementing iOffset each time
//
// The two-step approach is for speed since we only have to prepare the
// profile once for multiple iOffset's.  (We don't need the two-step
// approach for 2D profiles because we don't use whiskers for 2D profs and
// thus don't have to sample along a whisker).
//
// Shape is used for figuring out the direction of the whisker.  The whisker
// center point is x and y which may not be on the shape if there is an orthogonal
// offset.  iOrthOffset is offset from the point orthogonally to the whisker.
//
// Note on direction of "previous":
// if whisker is vertical, PrevPix is below this pixel

void
PrepareProf1D (const Image &Img, const Mat &Shape,                  // in: all
               unsigned SubProfSpec, const tLand LandTab[],
               int iPoint, int nProfWidth, double ProfAngle,
               int iOrthOffset, bool fExplicitPrevNext)
{
ASSERT((SubProfSpec & PROF_2d) == 0);
ASSERT(nProfWidth < MAX_PROF_WIDTH_1D);
ASSERT((SubProfSpec & PROF_TBits) == PROF_Grad);

ngProfWidth = nProfWidth;
igProfPoint = iPoint;       // for sanity checking in Get1dProf later
pgProfImage = Img.buf;      // ditto
gSubProfSpec = SubProfSpec; // ditto

// number of profile sample points
// nProfWidth is +-nSamplePoints and middle point
int nSamplePoints = (nProfWidth - 1) / 2;

double DeltaX;      // steps we move away along whisker from iPoint when sampling image
double DeltaY;      // DeltaX is along whisker, DeltaY is orthogonal to whisker
GetProfStepSize(DeltaX, DeltaY,
    Shape, iPoint, ProfAngle,LandTab, fExplicitPrevNext);

double x = GetX(Shape(iPoint, VX), 0, iOrthOffset, DeltaX, DeltaY);
double y = GetY(Shape(iPoint, VY), 0, iOrthOffset, DeltaX, DeltaY);

double PrevPix = iGetPixel(Img,
                    iRound(GetX(x, -nSamplePoints-1, 0, DeltaX, DeltaY)),
                    iRound(GetY(y, -nSamplePoints-1, 0, DeltaX, DeltaY)));

for (int iSample = -nSamplePoints; iSample <= nSamplePoints; iSample++)
    {
    int ix = iRound(GetX(x, iSample, 0, DeltaX, DeltaY));
    int iy = iRound(GetY(y, iSample, 0, DeltaX, DeltaY));
    double Pix = iGetPixel(Img, ix, iy);
    gProf(0, iSample + nSamplePoints) = Pix - PrevPix;
    PrevPix = Pix;
    }
}

//-----------------------------------------------------------------------------
static void
Normalize1d (Vec &Prof,             // io
             unsigned SubProfSpec)  // in
{
ASSERT((SubProfSpec & PROF_2d) == 0);
ASSERT((SubProfSpec & PROF_NormalizationField) == PROF_Flat);
const double Sum = Prof.absSum() / (Prof.ncols() * gNormalizedProfLen);
if (!fEqual(Sum, 0))
    Prof /= Sum;
}

//-----------------------------------------------------------------------------
// Use this after preparing global vars by calling PrepareProf1D.
// iOffset is offset along whisker from point, can be +ve or -ve.

void
Get1dProf (Vec &Prof,                                                   // out
     unsigned SubProfSpec, const Image &Img, int iPoint, int iOffset)   // in
{
ASSERT((SubProfSpec & PROF_2d) == 0);
ASSERT((SubProfSpec & PROF_FBit) == 0);
ASSERT(iPoint == igProfPoint);
ASSERT(Img.buf == pgProfImage);
ASSERT(SubProfSpec == gSubProfSpec);

const int nProfWidth = Prof.ncols();      // +-nSamplePoints and middle point
Prof = gProf.view(0,
            ngProfWidth/2 + iOffset - nProfWidth/2, 1, nProfWidth);

Normalize1d(Prof, SubProfSpec);
}

//-----------------------------------------------------------------------------
static void
Normalize2d (Vec &Prof,                                     // io
    const unsigned SubProfSpec, const double SigmoidScale)  // in
{
ASSERT(SubProfSpec & PROF_2d);
ASSERT((SubProfSpec & PROF_NormalizationField) == PROF_SigmAbsSum);
const int nelems = Prof.nelems();
int iCol = nelems;
double * const pData = Prof.m->data;    // for efficiency, access mat buf directly

// note that we scale but don't center here because
// we don't want to center an all positive profile

const double AbsSum = Prof.absSum();
const double Scale = nelems * gNormalizedProfLen / AbsSum;
while (--iCol >= 0)
    {
    pData[iCol] *= Scale;
    pData[iCol] /= (ABS(pData[iCol]) + SigmoidScale / 10);
    }
}

//-----------------------------------------------------------------------------
// Get the profiles from Grads.
// The Grads arrays must be initialized before calling this function.

static void
Get2dProfFromGradMats (Vec &Prof,                                              // out
    unsigned ProfSpec, int iSubProf, const MatVec &Grads, const SHAPE &Shape,  // in
    int iPoint, int ixOffset, int iyOffset, int nProfWidth)                    // in
{
ASSERT(IS_ODD(nProfWidth));
ASSERT(Grads[iSubProf].nrows() && Grads[iSubProf].ncols());
    // ensure gradient matrices are inited, done previously by InitGradsIfNeeded

unsigned SubProfSpec = GetSubProfSpec(ProfSpec, iSubProf);
ASSERT((SubProfSpec & PROF_WindowField) == PROF_WindowEquallyWeighted);

const int nProfWidthEachSide = (nProfWidth - 1) / 2;
const int nrows = Grads[iSubProf].nrows();
const int ncols = Grads[iSubProf].ncols();

// convert shape coords to mat coords,
// offset by ixOffset and iyOffset and nProfWidthEachSide

int iy = iRound(Shape(iPoint, VY)) - iyOffset;
int ix = iRound(Shape(iPoint, VX)) - ixOffset;

const int iyMin = nrows/2 - 1 - iy - nProfWidthEachSide;
const int iyMax = nrows/2 - 1 - iy + nProfWidthEachSide;
const int ixMin = ncols/2     + ix - nProfWidthEachSide;
const int ixMax = ncols/2     + ix + nProfWidthEachSide;

ASSERT(Prof.nelems() == unsigned(nProfWidth * nProfWidth)); // ensure we don't overrun Prof

double *pData = Prof.m->data;                           // for speed, access mat buf directly
const double * const pGrad = Grads[iSubProf].m->data;   // ditto
const int Tda = Grads[iSubProf].m->tda;

// For efficiency there are two loops: one to hande the case where the
// profile is entirely in the image boundaries and one for the case where
// the profile goes off the edge the edge of the image.

if (ixMin >= 0 && ixMax < ncols && iyMin >= 0 && iyMax < nrows) // in image boundary?
    for (iy = iyMin; iy <= iyMax; iy++)                 // yes, prof is in the image
        {
        const double *pGrad1 = pGrad + iy * Tda + ixMin;
        for (ix = ixMin; ix <= ixMax; ix++)
            *pData++ = *pGrad1++;
        }
else for (iy = iyMin; iy <= iyMax; iy++)                // prof crosses the boundary
    {
    const double *pGrad1 = pGrad + iy * Tda + ixMin;
    for (ix = ixMin; ix <= ixMax; ix++)
        {
        if (ix >= 0 && ix < ncols && iy >= 0 && iy < nrows)
            *pData++ = *pGrad1;
        else
            *pData++ = UNUSED_GRAD_VAL;
        pGrad1++;
        }
    }
}

//-----------------------------------------------------------------------------
void
Get2dProf (Vec &Prof,                                           // out
    const unsigned ProfSpec, const int iSubProf,                // in
    const MatVec &Grads,                                        // in
    const SHAPE &Shape, const int iPoint, const int ixOffset,   // in
    const int iyOffset, const int nProfWidth,                   // in
    const double SigmoidScale)                                  // in
{
ASSERT(ProfSpec & PROF_2d);
const unsigned SubProfSpec = GetSubProfSpec(ProfSpec, iSubProf);

Get2dProfFromGradMats(Prof, ProfSpec, iSubProf, Grads, Shape, iPoint,
        ixOffset, iyOffset, nProfWidth);

Normalize2d(Prof, SubProfSpec, SigmoidScale);
}

//-----------------------------------------------------------------------------
// After calling this function, m(i,j) corresponds to Img(j,i).
// Note the reversal of i,j.
//
// This means m.ncols() == Img.width
//            m.nrows() == Img.height
//
// And m.print() prints a matrix that must be rotated anticlockwise
// by 90 degrees to "look like" the corresponding image.

static void
ImageToMat (Mat &m,             // out
            const Image &Img)   // in
{
m.dim(Img.height, Img.width);

const int Tda = m.m->tda;

for (int iy = 0; iy < Img.height; iy++)
    {
    double *pData = m.m->data + iy * Tda;   // for speed, acces mat buf directly
    for (int ix = 0; ix < Img.width; ix++)
        *pData++ = Img(ix, iy);
    }
}

//-----------------------------------------------------------------------------
// Compare this pixel to the pixel on its right and pixel below it.
// Return sum (which can can be negative).
// This assumes that ix and iy are in range i.e. min ix is 0 and
// max ix is Img.width-2 (not Img.width-1).
// Ditto for iy.

static void
InitGrads_GradBelowRight (Mat &Grads,           // out
      const Mat &ImgMat)                        // in
{
const int ncols = ImgMat.ncols();
const int nrows = ImgMat.nrows();

for (int iy = 0; iy < nrows-1; iy++)
    for (int ix = 0; ix < ncols-1; ix++)
        {
        const double xDelta = ImgMat(iy,  ix+1) - ImgMat(iy,ix);
        const double yDelta = ImgMat(iy+1,ix)   - ImgMat(iy,ix);
        Grads(iy, ix) = xDelta + yDelta;
        }

// fill edges not covered in above for loops

Grads.fill(UNUSED_GRAD_VAL, nrows-1, 0,       nrows, ncols);
Grads.fill(UNUSED_GRAD_VAL, 0,       ncols-1, nrows, ncols);
}

//-----------------------------------------------------------------------------
// Given an image Img, generate the gradient arrays Grads for the image.
// ProfSpec tells us what gradients to generate.  We generate grads only
// for 2D profiles.

static void
InitGrads (MatVec &Grads,                           // out
           const Image &Img, unsigned ProfSpec)     // in
{
const int ncols = Img.width;
const int nrows = Img.height;
const int nSubProfs = nSubProfsForProfSpec(ProfSpec);
ASSERT(nSubProfs == 1);
Grads.resize(nSubProfs);

// Convert the image to matrix form so no type conversions needed below (for speed).
// Remember that Img(i,j) == ImgMat(j,i)   (indices are swapped).

Mat ImgMat;
ImageToMat(ImgMat, Img);

ASSERT(MAX_NBR_SUBPROFS == 1);  // this func supports only one subprof per prof
const unsigned iSub = 0;        // ...so this index is always 0
unsigned SubProfSpec = GetSubProfSpec(ProfSpec, iSub);
if (SubProfSpec & PROF_2d)
    {
    ASSERT((SubProfSpec & PROF_FBit) == 0);
    ASSERT((SubProfSpec & PROF_TBits) == PROF_GradBelowRight);
    Grads[iSub].dim(nrows, ncols);
    InitGrads_GradBelowRight(Grads[iSub], ImgMat);
    }
}

//-----------------------------------------------------------------------------
// For now, all the 2D profile types must be the same because we store
// Grads on an image basis, not on a landmark basis.
// So here we check that they do indeed all match.
//
// Returns the first 2d ProfSpec in ProfSpecs, if any.

static unsigned
CheckThatAll2dProfsMatch (const cvec<unsigned> &ProfSpecs, int nPoints) // in: all
{
unsigned ProfSpec = 0;
for (int iPoint = 0; iPoint < nPoints; iPoint++)
    {
    unsigned ProfSpec1 = ProfSpecs[iPoint];
    if (ProfSpec1 & PROF_2d)
        {
        if (ProfSpec == 0)
            ProfSpec = ProfSpec1;
        else if (ProfSpec1 != ProfSpec)
            Err("ProfSpec not constant over all 2D profiles");
        }
    }
return ProfSpec;
}

//-----------------------------------------------------------------------------
void
InitGradsIfNeeded (MatVec &Grads,       // out
    const cvec<unsigned> &ProfSpecs,    // in: prof specs read from ASM file
    const Image &Img,                   // in: Img already scaled to this pyr level
    int nPoints)                        // in:
{
const unsigned ProfSpec = CheckThatAll2dProfsMatch(ProfSpecs, nPoints);

if (ProfSpec & PROF_2d)
    InitGrads(Grads, Img, ProfSpec);
}
