// $startshape.cpp 1.5 milbo$ routines for finding start shape for an ASM search
//-----------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// A copy of the GNU General Public License is available at
// http://www.r-project.org/Licenses/
//-----------------------------------------------------------------------------

#include "all.hpp"

#if _MSC_VER    // microsoft
#define VIOLA_JONES 1
#else
#define VIOLA_JONES 0   // viola jones detector is currently only supported under MSC
                        // (but only because I don't yet have the VJ libs for mingw)
#endif

// following are for aligning base shape to Rowley detector shape

static const double CONF_RowleyHeightRatio  = 0.60;
static const double CONF_RowleyWidthRatio   = 0.07;
static const double CONF_RowleyEyeWeight    = 1.0;
static const double CONF_RowleyWidthWeight  = 0.5;
static const double CONF_RowleyHeightWeight = 0.5;
static const double CONF_MaxEyeToFaceWidth  = 0.52;
static const double CONF_EyeFaceBlend       = 0.60;

// following are for aligning base shape to Viola Jones detector shape

static const double CONF_VjHeightShift = 0.15;  // shift height down by 15%
static const double CONF_VjShrink      = 0.8;   // shrink size of VJ box by 20%

//-----------------------------------------------------------------------------
// Make the Rowley Jones face box smaller and move it down a bit.
// This gives a better position for the start shape.

#if VIOLA_JONES
static void AdjustViolaJonesShape (SHAPE &Shape) // io
{
double xMin = Shape(DETECTOR_TopLeft, VX);
double yMin = Shape(DETECTOR_TopLeft, VY);
double xMax = Shape(DETECTOR_BotRight, VX);
double yMax = Shape(DETECTOR_BotRight, VY);

double NewHeight = CONF_VjShrink * (yMax - yMin);
double yMean =
    (Shape(DETECTOR_TopLeft, VY) + Shape(DETECTOR_BotRight, VY)) / 2;
yMean += CONF_VjHeightShift * (yMax - yMin);    // move face down

double NewWidth = CONF_VjShrink * (xMax - xMin);
double xMean =
    (Shape(DETECTOR_TopLeft, VX) + Shape(DETECTOR_BotRight, VX)) / 2;

Shape(DETECTOR_TopLeft,  VX) = xMean - .5 * NewWidth;
Shape(DETECTOR_TopLeft,  VY) = yMean - .5 * NewHeight;
Shape(DETECTOR_BotRight, VX) = xMean + .5 * NewWidth;
Shape(DETECTOR_BotRight, VY) = yMean + .5 * NewHeight;
}
#endif

//-----------------------------------------------------------------------------
// align MeanShape to the Viola Jones global detector face box

#if VIOLA_JONES
static void
AlignToViolaJones (SHAPE &StartShape,                   // out
        const SHAPE &MeanShape, const SHAPE &DetShape)  // in
{
SHAPE Base(2, 2), VjAlign(2, 2);   // MlJaw2 MRJaw2

Base.row(0) = MeanShape.row(MLJaw2);
Base.row(1) = MeanShape.row(MRJaw2);

double yMean = (DetShape(DETECTOR_TopLeft, VY) + DetShape(DETECTOR_BotRight, VY)) / 2;
VjAlign(0, VX) = DetShape(DETECTOR_TopLeft, VX);
VjAlign(0, VY) = yMean;
VjAlign(1, VX) = DetShape(DETECTOR_BotRight, VX);
VjAlign(1, VY) = yMean;

ASSERT(!Base.fARowIsZero());
ASSERT(!VjAlign.fARowIsZero());

Mat Pose(AlignShape(Base, VjAlign, NULL));

StartShape = TransformShape(MeanShape, Pose);
}
#endif

//-----------------------------------------------------------------------------
// align MeanShape to the Rowley detector eyes (ignore the overall face box)

static void
AlignToRowleyEyes (SHAPE &StartShape,                   // out
        const SHAPE &MeanShape, const SHAPE &DetShape)  // in
{
SHAPE Base(2, 2), Rowley(2, 2);   // points are MLEye MrEye

ASSERT(MLEye < int(MeanShape.nrows()) && MREye < int(MeanShape.nrows()));

Base.row(0) = MeanShape.row(MLEye);
Base.row(1) = MeanShape.row(MREye);

// in practice the Rowley detector sometimes misaligns the eyes so here
// it's better to just use the average height of the eyes

double yMean = (DetShape(DETECTOR_LEye, VY) + DetShape(DETECTOR_REye, VY)) / 2;
Rowley(0, VX) = DetShape(DETECTOR_LEye, VX);
Rowley(0, VY) = yMean;
Rowley(1, VX) = DetShape(DETECTOR_REye, VX);
Rowley(1, VY) = yMean;

ASSERT(!Base.fARowIsZero());
ASSERT(!Rowley.fARowIsZero());

Mat Pose(AlignShape(Base, Rowley, NULL));

StartShape = TransformShape(MeanShape, Pose);
}

//-----------------------------------------------------------------------------
// align MeanShape to the Rowley detector overall face box and the eyes

static void
AlignToRowleyFaceAndEyes (SHAPE &StartShape,            // out
        const SHAPE &MeanShape, const SHAPE &DetShape)  // in
{
SHAPE Base(4, 2), Rowley(4, 2);   // points are MLEye MrEye MlJaw2 MRJaw2

ASSERT(MLEye < int(MeanShape.nrows()) && MREye < int(MeanShape.nrows()));

Base.row(0) = MeanShape.row(MLEye);
Base.row(1) = MeanShape.row(MREye);
Base.row(2) = MeanShape.row(MLJaw2);
Base.row(3) = MeanShape.row(MRJaw2);

ASSERT(!Base.fARowIsZero());

// in practice the Rowley detector sometimes misaligns the eyes so here
// it's better to just use the average height of the eyes

double yMean = (DetShape(DETECTOR_LEye, VY) + DetShape(DETECTOR_REye, VY)) / 2;
Rowley(0, VX) = DetShape(DETECTOR_LEye, VX);
Rowley(0, VY) = yMean;
Rowley(1, VX) = DetShape(DETECTOR_REye, VX);
Rowley(1, VY) = yMean;

double xMin = DetShape(DETECTOR_TopLeft, VX);
double xMax = DetShape(DETECTOR_BotRight, VX);

// recenter the face on the midpoint of eyes (found to give better results)

double EyeMidpoint = (DetShape(DETECTOR_LEye, VX)+DetShape(DETECTOR_REye, VX))/2;
xMin = EyeMidpoint - (xMax - xMin)/2;
xMax = EyeMidpoint + (xMax - xMin)/2;

double yMin = DetShape(DETECTOR_TopLeft, VY);
double yMax = DetShape(DETECTOR_BotRight, VY);

// y position of Jaw2 in Rowley box
double HeightCorrect = CONF_RowleyHeightRatio * (yMax - yMin);

// difference between MLJaw2-to-MRJaw2 and Rowley box
double WidthCorrect  = CONF_RowleyWidthRatio * (xMax - xMin);

Rowley(2, VX) = xMin + WidthCorrect;   Rowley(2, VY) = yMin + HeightCorrect;
Rowley(3, VX) = xMax - WidthCorrect;   Rowley(3, VY) = yMin + HeightCorrect;

ASSERT(!Rowley.fARowIsZero());

Vec Weights(Base.nrows());
Weights(0) = CONF_RowleyEyeWeight;
Weights(1) = CONF_RowleyEyeWeight;
Weights(2) = CONF_RowleyWidthWeight;
Weights(3) = CONF_RowleyHeightWeight;

Mat Pose(AlignShape(Base, Rowley, &Weights));

StartShape = TransformShape(MeanShape, Pose);
}

//-----------------------------------------------------------------------------
static void
AlignToRowley (SHAPE &StartShape,                           // out
            const SHAPE &MeanShape, const SHAPE &DetShape)  // in
{
AlignToRowleyFaceAndEyes(StartShape, MeanShape, DetShape);

// Align the MeanShape using only the Rowley eye positions, to give
// the shape StartEyes.
// Measure the eye positions in StartEyes against those in the StartShape
// we got above (which was aligned to both the detector face and detector eyes).
// If the eye positions are very different, then form a new StartShape by blending
// StartShape and StartEyes.
//
// This has proven effective in improving the fit in faces with large alignment
// errors, without affecting other faces.

SHAPE StartEyes;
AlignToRowleyEyes(StartEyes, MeanShape, DetShape);

double EyeToFaceWidth =
    (StartEyes(MREye, VX) - StartEyes(MLEye, VX)) /
    (StartShape(MRJaw2, VX) - StartShape(MLJaw2, VX));

double Blend = 0;   // assume no blending
if (EyeToFaceWidth > CONF_MaxEyeToFaceWidth)
    Blend = CONF_EyeFaceBlend;

unsigned i;
for (i = 0; i <= MNoseTip; i++)
    StartShape.row(i) = (1-Blend) * StartShape.row(i) + Blend * StartEyes.row(i);

for (i = MLEye0; i < StartShape.nrows(); i++)   // extra eye points
    StartShape.row(i) = (1-Blend) * StartShape.row(i) + Blend * StartEyes.row(i);
}

//-----------------------------------------------------------------------------
// Align the model MeanShape to a global detector shape to provide a
// good starting point for the search.  Results are returned in StartShape.

void
GetStartShapeFromGlobalDet (SHAPE &StartShape,          // out
                            bool fViolaJones,           // in
                            Image &Img,                 // in
                            const char sImageFile[],    // in
                            const SHAPE &MeanShape,     // in
                            const char sDataDir[])      // in
{
SHAPE DetShape;   // see DETECTOR_ defs in landmarks.hpp for format

if (fViolaJones)
#if VIOLA_JONES
    {
    ViolaJonesFindFace(DetShape, Img, sImageFile, sDataDir);
    AdjustViolaJonesShape(DetShape);
    AlignToViolaJones(StartShape, MeanShape, DetShape);
    }
#else
    Err("Viola Jones detector not supported");
#endif
else    // Rowley detector
    {
    RowleyFindFace(DetShape, Img, sImageFile, sDataDir);
    AlignToRowley(StartShape, MeanShape, DetShape);
    }
// An all zero row means both the x and y coords are zero.
// But x and y both equal to 0 means the point is "unused".
// So assert that that is not the case.
// TBD should do this properly by setting all 0 row to a small nbr.

ASSERT(!StartShape.fARowIsZero());
}


//-----------------------------------------------------------------------------
void GetStartShapeFromPreviousSearch (SHAPE &StartShape,     // out
        const SHAPE &PrevShape,     // in: combined shape from previous search
        const SHAPE &MeanShape)     // in
{
if (MeanShape.nrows() == PrevShape.nrows())
    StartShape = PrevShape;
else if (MeanShape.nrows() < PrevShape.nrows())
    {
    StartShape = PrevShape;
    StartShape.dimKeep(MeanShape.nrows(), 2);
    }
else    // MeanShape.nrows() > PrevShape.nrows()
    {
    // The algorithm below is: Use points from PrevShape where possible.  Where a point
    // is unused in PrevShape, use the point from the aligned MeanShape instead.

    StartShape.assign(MeanShape);  // use "assign" not "=" because dims may be different
    SHAPE Combined1(PrevShape);
    Combined1.dimKeep(MeanShape.nrows(), 2);
    AlignShape(StartShape, Combined1);
    for (unsigned iRow = 0; iRow < PrevShape.nrows(); iRow++)
        if (fPointUsed(StartShape, iRow))
            StartShape.row(iRow) = PrevShape.row(iRow);
    }
SetUnusedPointsToZero(StartShape, MeanShape);
}
