// $util.cpp 1.5 milbo$ utilities for stasm.cpp
//
// If you are linking the "stasm" code into an existing application you
// will probably have to tweak some routines in this file.
//-----------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// A copy of the GNU General Public License is available at
// http://www.r-project.org/Licenses/
//-----------------------------------------------------------------------------

#include "all.hpp"

FILE *pgLogFile;                        // log file: lprintfs go here
static const int MAX_PRINT_LEN = 10000; // big enough for big printfs
static char sgBuf[MAX_PRINT_LEN];       // general purpose buffer

//-----------------------------------------------------------------------------
// After calling this, on program termination we will get a report if there are
// writes outside the borders of allocated blocks or if there are non-freed blocks.

#if _MSC_VER && _DEBUG // microsoft and debugging enabled
static void InitMallocDebug (void)
{
_CrtSetReportMode(_CRT_WARN, _CRTDBG_MODE_WNDW);
_CrtSetReportMode(_CRT_WARN, _CRTDBG_MODE_FILE);
_CrtSetReportFile(_CRT_WARN, _CRTDBG_FILE_STDOUT);

// not included in Flag: _CRTDBG_CHECK_ALWAYS_DF _CRTDBG_CHECK_CRT_DF
int Flag = _CrtSetDbgFlag(_CRTDBG_REPORT_FLAG);
Flag |= (_CRTDBG_ALLOC_MEM_DF|_CRTDBG_DELAY_FREE_MEM_DF|_CRTDBG_LEAK_CHECK_DF);
_CrtSetDbgFlag(Flag);
}
#endif

//-----------------------------------------------------------------------------
// This function is called from within malloc or new if they fail.
// See the entry for _set_new_handler in the Microsoft VC++ documentation.
// We set it up by allocating a memory block for the printf function to use
// in case of memory allocation failure; the printf function uses malloc so we
// have to keep some memory in reserve so printf can be used when memory runs out.

#if _MSC_VER    // microsoft

char *pgFailsafe = new char[512];

static int __cdecl HandleProgramMemoryDepletion (size_t size)
{
delete pgFailsafe;  // release memory held in reserve
lprintf("\nOut of memory allocating %g megabytes\n", double(size) / (1024 * 1024));
exit(1);
}
#endif // microsoft

//-----------------------------------------------------------------------------
// Assign the function that will be called if malloc or new fails.
// Note that gcc uses a different method for checking malloc fails: see
// the definitions of _malloc and _calloc in misc.hpp.

#if _MSC_VER // microsoft
void SetMallocHandler (void)
{
    _set_new_handler(HandleProgramMemoryDepletion);
    _set_new_mode(1);
    #if _DEBUG // debugging enabled?
    InitMallocDebug();
    #endif
}
#else        // gcc
void *_malloc (size_t size)
{
void *p = malloc(size);
if (!p)
    {
    lprintf("\nOut of memory allocating %g megabytes\n",
        double(size) / (1024 * 1024));
    exit(1);
    }
return p;
}

void *_calloc (size_t num, size_t size)
{
void *p = calloc(num, size);
if (!p)
    {
    lprintf("\nOut of memory allocating %g megabytes\n",
        double(num * size) / (1024 * 1024));
    exit(1);
    }
return p;
}
#endif // microsoft

//-----------------------------------------------------------------------------
// called by the GSL routines if there is an error

void GslErrorHandler (const char * reason, const char * file, int line, int gsl_errno)
{
Err("%s:%d %s errno %d", file, line, reason, gsl_errno);
}

//-----------------------------------------------------------------------------
// Shutdown is called when stasm finishes (either correctly or with an error)

void Shutdown(void)
{
fclose(pgLogFile);
FreeNetworks();
#if _MSC_VER    // microsoft
CloseViolaJones();
if (pgFailsafe)
    delete(pgFailsafe);
#endif
}

//-----------------------------------------------------------------------------
// Opens the log file. After this when you call lprintf, you print on the
// screen and the log file.

FILE *pOpenLogFile (const char sLogFile[])
{
pgLogFile = fopen(sLogFile, "w");
if (!pgLogFile)
    {
    printf("\nCan't open \"%s\"\n", sLogFile);
    exit(1);    // we don't call Err() to avoid possible recursive Err
    }
return pgLogFile;
}

//-----------------------------------------------------------------------------
// like printf but prints to the log file as well

void lprintf (const char *pArgs, ...)   // args like printf
{
va_list pArg;
va_start(pArg, pArgs);
vsprintf(sgBuf, pArgs, pArg);
va_end(pArg);
printf("%s", sgBuf);
fflush(stdout);     // flush so if there is a crash we can see what happened
if (pgLogFile)
    fputs(sgBuf, pgLogFile);
}

//-----------------------------------------------------------------------------
// Print an error message and exit.
// This puts a \n on the msg so you shouldn't.

void Err (const char *pArgs, ...)       // args like printf
{
static bool fInHere;
if (fInHere)            // if already in here then just exit
    printf("\nRecursive call to Err\n");
else
    {
    fInHere = true;
    va_list pArg;
    va_start(pArg, pArgs);
    vsprintf(sgBuf, pArgs, pArg);
    va_end(pArg);
    printf("\nError: %s\n", sgBuf);
    fflush(stdout);
    if (pgLogFile)
        {
        fprintf(pgLogFile, "\nError: %s\n", sgBuf);
        fflush(pgLogFile);
        }
#if _DEBUG
    if (ENTER_DEBUGGER_ON_ERR)
        ENTER_DEBUGGER;
#endif
    Shutdown();
    }
exit(1);
}

//-----------------------------------------------------------------------------
// like fopen but prints a message and exits if the file open fails

FILE *Fopen (const char *sFile, const char *sMode)
{
FILE *pFile = fopen(sFile, sMode);
if (!pFile)
    {
    const char *s;
    switch (sMode[0])
        {
        case 'w':
            s = " for writing";
            break;
        case 'a':
            s = " for appending";
            break;
        case 'r':
            s = "";
            break;
        default:
            s = " (unknown mode)";
            break;
        }
    Err("Can't open %s%s", sFile, s);
    }
return pFile;
}

//-----------------------------------------------------------------------------
// like fwrite but prints a message and exits if write fails

void Fwrite (const void *buf, size_t Size, size_t Count, FILE *pFile, char *sFile)
{
if (fwrite(buf, Size, Count, pFile) != Count)
    Err("Can't write %s", sFile);
}

//-----------------------------------------------------------------------------
// like fprintf but prints a message and exits if fprintf fails

void Fprintf (FILE *pFile, const char *pArgs, ...)
{
va_list     pArg;
va_start(pArg, pArgs);
vsprintf(sgBuf, pArgs, pArg);
va_end(pArg);
int Len = strlen(sgBuf);
ASSERT(Len < MAX_PRINT_LEN);
if (Len != fprintf(pFile, "%s", sgBuf))
    Err("Can't write to file: %s", sgBuf);
}

//-----------------------------------------------------------------------------
// like Fprintf above but has a filename parameter for better error reporting

void Fprintf (const char *sFile, FILE *pFile, const char *pArgs, ...)
{
va_list     pArg;
va_start(pArg, pArgs);
vsprintf(sgBuf, pArgs, pArg);
va_end(pArg);
int Len = strlen(sgBuf);
ASSERT(Len < MAX_PRINT_LEN-1);
if (Len != fprintf(pFile, "%s", sgBuf))
    Err("Can't write to %s: %s", sFile, sgBuf);
}

//-----------------------------------------------------------------------------
// like fread but prints a message and exits if the read fails

void Fread (void *buf, size_t Size, size_t Count, FILE *pFile, const char *sFile)
{
if (fread(buf, Size, Count, pFile) != Count)
    Err("Can't read %s", sFile);
}

//-----------------------------------------------------------------------------
// like gets but skips comments i.e. white space lines and lines beginning with #

char *Fgets (char *s, int n, FILE *stream)
{
char *p;
do
    p = fgets(s, n, stream);
while ((p != NULL) && (s[0] == '#' || strspn(s, " \t\n") == strlen(s)));
return p;
}
